/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Wang Ke
 */

#ifndef ADINS_H
#define ADINS_H

#include <stdlib.h>
#include <math.h>
#include "stream.h"
#include "virtstream.hpp"

// macros to ease code writing
#define AMIN             Config::min_ad_time
#define AMAX             Config::max_ad_time
#define VMIN             Config::min_video_time
#define TPRIME           Config::long_term_ad_time
#define MAXT             Config::long_term_window
#define R_VMIN_AMIN      Config::ratio_vmin_amin
#define R_MAX_MIN_AD     Config::ratio_max_min_ad
#define R_TPRIME_AMAX    Config::ratio_tprime_amax
#define MAX_VMIN         (Config::long_term_video)
#define LAST_VIDEO_TIME  (Config::long_term_video - \
                         (Config::ratio_tprime_amax - 1)*Config::min_video_time)


int compare_streams(const void *, const void *);
int compare_clusters(const void *, const void *);

template<class T>
inline const T& max(const T& a, const T& b) { return ((a>b) ? a:b);}

template<class T>
inline const T& min(const T& a, const T& b) { return ( (a>b)? b:a);}

inline int isodd(int i) { return ( ((i%2)==1)? 1:0);}

inline double modulus(double a, double b) { return( (a - floor(a/b)*b));}

template<class T>
inline int iszero(T i) { return ( (i==(T)0)? 1:0);}

template<class T>
inline T abs(T i) {return ( (i >= (T)0) ? i:-(i));}

const int MAX_ARRAY_STTS = 3;
const int FIRST = 0;
const int ADI_SUCCESS = 1;
const int ADI_FAILURE = -1;

const int ZERO_VALUE = 0;
const int UNINITIALIZED = -1;
const int LEAF = -1;
const int NOT_FOUND = -1;
const double STREAM_NOT_FOUND = -2.0;

// Cluster class to group information together
class Cluster {
public:
  static int num_cluster;
  int first_member; // leading stream
  int last_member;

  Cluster():first_member(UNINITIALIZED), last_member(UNINITIALIZED) {num_cluster++;};
  ~Cluster() {num_cluster--;};
  void printData() { cout << "first: " << first_member << "\nlast: " 
			  << last_member << endl;}
};


// Command class to group information together
class Command {
public:
  static int num_command;
  Status status;
  double position_after_command;
  double duration;

  Command():status(UNKNOWN), position_after_command(0.0), duration(0.0) \
    {num_command++;};
  ~Command() {num_command--;};
};


// PositionNode - used to build optimal tree
class PositionNode {
public:
  static int num_pn;
  double position;
  double ad_to_go;
  double time_to_go;
  double video_to_go;
  PositionNode():position(0.0),ad_to_go(0.0),time_to_go(0.0), video_to_go(0.0)\
    {num_pn++;}
  ~PositionNode() {num_pn--;};
};


// TriangMatrix is used to store the information of the binary tree
template<class T> 
class TriangMatrix {
private:
  void init_vec(int mne);

public:
  T* elem;
  int side;
  int max_side;
  int max_num_elem;
  static int num_tmatrix;
  TriangMatrix();
  ~TriangMatrix();

  void init_matrix(int mxsd);
  void deallocate_matrix();
  void reset_matrix(int newside);
  T get_value(int i, int j);
  T& get_obj(int i, int j);
  void set_value(int i, int j, T t);
};


// The Stream class stores all the information of the beginning of the
//  snapshot and the end information of the snapshot
class Stream {

public:
  const static int max_ncommands = 12;
  static int num_stream;
  double stream_id;
  Status status;

  Command commands[max_ncommands];
  int last_command_prepended;

  int movie_id;
  double movie_length;

  double position; 

  double ad_to_go;   // amount of ad time that is left 
  double time_to_go; // amount of ad time left in the long window interval
  double video_to_go; 
  // amount of video time left in the long window interval

  double cost;
  double merging_with;
  double merging_point;

  Stream();
  ~Stream();

  int init(double str_id, double pos, int movie, double movie_len,
	   Status status, double atg, double ttg, double vtg,
	   double window);

  int init(STREAM *sts, double window);
  void prepend_command(Status stts, double till_position);
  Status map_status(NStatus st);
  double get_cost();
  double get_original_cost();
  void printData();
  void printData(int i);
  void printCommands();

private:
  double cost_per_command(Command *cmd);

};


// Class that performs the clustering task on the streams
class Clusterer {

protected:
  void sort_streams(Stream *streams, int nstreams);

public:
  static int num_clusterer;
  Clusterer();
  virtual ~Clusterer();
  PositionNode converge_at(Stream *s1, Stream *s2, double interval, int clust);
  virtual void cluster(Stream *streams, int nstreams, 
		       Cluster *clusters, double window,
		       int *nclusters, int *nchannels);
};


// Specific EMCL clustering
class EMCL_Clusterer:public Clusterer{

public:
  static int num_emcl;
  EMCL_Clusterer();
  virtual ~EMCL_Clusterer();

  // stream s1 must be "ahead" of stream s2 =>  s1 is leading, s2 is trailing
  //  double can_converge(Stream *s1, Stream *s2, double window);
  void cluster(Stream *streams, int nstreams,
	      Cluster *clusters, double window,
	      int *nclusters, int *nchannels);
};

// Class that performs the Ad insertion computations on the streams
class Ad_Insertion {

  Stream *streams;
  Cluster *clusters;
  Clusterer *clusterer;

  TriangMatrix<double> cost_tree;
  TriangMatrix<PositionNode> position_tree;
  TriangMatrix<int> t_tree;

  int nstreams;
  int nclusters;
  int nchannels;
  int max_cluster_element;

public:
  double TIMEBASE;

  Ad_Insertion(Clusterer *clust);
  ~Ad_Insertion();

  NStatus get_stream_status(double stream_id, 
			    double position,
			    double time);
  
  void invoke_policy(STREAM *streams,
		     int num_streams,
		     double window,
		     double current_time);
  double get_cost();
  double get_original_cost();
  double cost_for_commands();
  void debug_streams();
  void debug_streams(int i, int j);
  void debug_stream_id(double id);
  void debug_chain_id(double start);
  void debug_clusters();
  void print_stream_cmd();
  void debug_tree_struct();
  void debug_tree_struct(int which_cluster);
  void debug_tree_struct(int leading, int trailing, int first);
  void checkTriangMatrix();
  double debug_merge_with(double);
  double to_merge_with(double);

private:
  double compute_amount_ad_needed(int,int,double);
  void allocate(STREAM *sts, int nstreams, double window);
  void deallocate();
  void set_one_cluster(int which, double window);

  // i is the leading stream, j is the trailing stream
  void compute_merge_point(int i, int j, int first_stream, double window, int clust);
  void set_min_cost_tree(int i, int j, int first_stream, double movie_length);
  void build_paths(int leading, int trailing, int first_stream, int last_merge);

  void build_stream_cmd_path(int offset, int first_stream);
  int find_stream(double stream_id);

  // find a not merged stream inside cluster which_cluster, starting at
  //  start_position
  NStatus get_status_from_command(Status stts, double offset);
  double get_position_at_time(int str_index, int cmd_index, 
			      double offset);
  void assign2(Stream *lead, Stream *trail, PositionNode *p, int last_merge);

};
#endif
