/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Wang Ke
 */

#ifndef __CHANNEL_HPP__
#define __CHANNEL_HPP__

#include "virtstream.hpp"
#include <assert.h>

enum ChannelType {Undef, OnDem, Cont, Both};

class Channel:public Node
{
friend class Channellist;
private:
  Virtualstream *v;
  int channelID;
  ChannelType channelType;
public:
  static int num_channel;
  static int id_counter;

  // variable to keep track of the time the channel is assigned
  // assigned to -1 on constructor and on ChannelManager::returnActiveChannel
  // assigned to k on ChannelManager::getIdleChannel
  double assignedAt;

  Channel() {channelID = ++id_counter; num_channel++; channelType = Undef;\
             assignedAt = -1.0;};
  ~Channel() {num_channel--;};
  int Isbusy(void) {return (v ? 1:0);};
  void setVS(Virtualstream *vs) { v = vs;}; 
  void Dumpdata(void);
  int getID() {return channelID;};
  void setType(ChannelType type) {channelType = type;};
  ChannelType getType() {return channelType;};
  Channel *getNext() {return (Channel *)next;};
  Channel *getPrev() {return (Channel *)prev;};
};

class Ad_Channel: public Channel
{
protected:
  double last_ad_start_time;
  double ad_duration;

public:
  static int num_ad_channel;
  Ad_Channel(double ad_dur) { ad_duration = ad_dur; num_ad_channel++; };
  ~Ad_Channel() {num_ad_channel--;};
  void set_last_ad_time(double time) { last_ad_start_time = time; }
  double get_last_ad_time() { return last_ad_start_time; }
  double get_nearest_ad_time(double timenow);
};


class Channellist:public List
{
public:
  static int num_channellist;
  Channellist();
  ~Channellist();

  void buildList(int n);
  Channel *detachChannel();
  Channel *detachChannel(int which);
  //  void appendChannel(Channel *); - use Prepend(Channel *, NULL)
};

class QueueManager;

class ChannelManager
{
public:
  static int num_channelmanager;
  ChannelManager(int, double);
  ~ChannelManager();
  
  double fractionOnDemand;
  int numOnDemand;
  int numContingency;
  int totNumber;

  QueueManager *qmgr;

  Channellist activeOnDemandlist;
  Channellist idlelist;
  Channellist activeContingencylist;

  Channel *getIdleChannel(ChannelType);
  void returnActiveChannel(Channel *);
  void putToActive(Channel *);

  int onDemandAvailable();
  void waitForOnDemandChannel(Virtualstream *);

  int contingencyAvailable();
  void waitForContingencyChannel(Virtualstream *);


  void serveOnDemandRequests(VSList *bmv);
  void serveContingencyRequests(VSList *rmv);
  void checkWaitUntilLimits(Subscriberlist *active, Subscriberlist *idle);

  void changeFraction(double);
};

class WaitStat
{
public:
  static int num_waitstat;

  int stream_waiting;
  double total_wait_time;
  double total_num_wait;
  double upThreshold;
  double moreUp;
  double lowThreshold;
  double lessLow;
  double tot_quitters;
  double tot_quit_wt;

  WaitStat() {num_waitstat++; stream_waiting = 0;};
  ~WaitStat() {num_waitstat--;};
};

class Waitlist:public List
{
public:
  static int num_waitlist;

  WaitStat wstat;

  Waitlist() {num_waitlist++;};
  ~Waitlist() {DeleteAllNodes(); num_waitlist--;};
  void appendVirtualstream(Virtualstream *);
  Virtualstream *detachVirtualstream();
  double getAverageWT();
  int joinVS(Virtualstream *, ChannelManager *);
  void compact(Subscriberlist *active, Subscriberlist *idle);
};

class QueueManager
{
public:
  static int num_queuemanager;
  QueueManager() {num_queuemanager++;};
  ~QueueManager() {num_queuemanager--;};

  double upThreshOnDemand;
  double lowThreshOnDemand;
  double upThreshContingency;
  double lowThreshContingency;
  Subscriberlist *sl;

  Waitlist onDemandlist;
  Waitlist contingencylist;

  double getAverageWT(ChannelType which);
  double getPercUp(ChannelType which);
  double getPercLow(ChannelType which);
  double getUp(ChannelType which);
  double getLow(ChannelType which);
  double getTotal(ChannelType which);
  double getQuit(ChannelType which);
  double getQuitWT(ChannelType which);
};

#endif

