/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan, 
 * Prithwish Basu,
 * Wang Ke
 */
 
#ifndef __VIRTSTREAM_HPP__
#define __VIRTSTREAM_HPP__


// Batching Policy
#define TIMEOUT_POLICY 0 
#define EPSILON 0.0022


#include "subscriber.hpp"
#include "movie.hpp"
#include "stream.h"
#include "list.hpp"

class Cost
{
public:
  static int num_cost;
  double original;
  double used;
  double amount_in_ad;
  double amount_in_play;
  Cost() {original = used = amount_in_ad = amount_in_play = 0.0;num_cost++;};
  ~Cost() {num_cost--;};
};


enum Status {RTOPLAY, PLAYING, FASTPLAY, AD_INS, INBATCH,PAUSED, FWDING, RWDING, 
	     ATEOMV, ADVACN,
	     ADINS_STT, PLAYING_STT, MERGED_STT, TRAILING_STT, 
	     ADI_ADINS_COM, ADI_ADONLY_COM, ADI_PLAYING_COM, ADI_VIDEO_AD_COM,
	     PL_PLAYING_COM, 
	     UNKNOWN};

//enum Status {RTOPLAY, PLAYING, FASTPLAY, PAUSED, FWDING, RWDING, 
//	     ATEOMV, INBATCH, ADVACN, UNKNOWN};

class Channel;
class WaitStat;
class ChannelManager;
class MStream;

class Virtualstream:public Node
{
  friend class VSList;
private:
  double stream_id;
  Channel *chan;
  Movie *mv;
  double position;
  Status status;
  double until_time;
  List users; 	
public:

  // s_video_to_go keeps track of the "short term video time" needed after 
  // ad time. video_to_go keeps track of the total amount of video time the
  // stream needs in a long term window. After the needed amount is reached, 
  // video_to_go is changed to a negative value  
  // to indicate that the stream need not give ads anymore (leave ADINS_STT
  //  state) but just display video (in PLAYING_STT state)

  static int num_vs;

  // after_interaction is set at Virtualstrea::Advancestream, for all
  // interacting states. It is reset either at Joinstream (if there is 
  // a non-after_interaction stream) or at Aggregate with recomp == 1.
  // been_interaction is set at AdStrList->Dispatch or at GENERATE_VCR_ACTIONS
  // function calls. It is used mainly for bogus stream purposes. It is
  // reset at Aggregate with recomp == 1. 

  char after_interaction;
  char been_interaction;

  // If the stream is a bogus stream. Bogus streams carry
  // the statistics of streams which interacted. 
  // They are necessary because QoS statistics are carried by each
  // stream and these are updated during Joinstream calls.
  // Streams which interact might not join streams which they had been
  // scheduled to and thus the QoS regulations are violated.

  char bogus;

  // newArrival is set when the streams first arrives at the VoD system
  // it is used for computing statistics purpose and should not
  // influence any system behavior. Set at BATCH_REQUESTS and
  // reset by the first call to Aggregate with recomp == 1

  char newArrival;

  // first_merge is set at Aggregate with recomp == 1
  // and reset at Joinstream if either stream has first_merge == 1.
  // It is also reset at constructor.
  // This is used to track how long a stream takes before merging with
  // another stream after the recomputation process took place. And
  // last_recomp_interval is set to the recomputation time unless
  // after_recomp is still 1 (i.e., the stream did not merge 
  char first_merge;
  double last_recomp;


  // stream id of the stream the clustering engine scheduled
  // for the stream to merge with.
  double to_merge_with;
  double ad_to_go;
  double time_to_go;
  double s_video_to_go;
  double video_to_go;
  double continuous_video_time;
  int n_ad_time;
  int n_batched;
  MStream *mstr;
  List *mstrlist;

  Channel *GetChan() {return chan;};
  void SetChan(Channel *c);
  Movie *GetMv() { return mv;};
  double GetUntil() { return until_time;};
  double GetMvduration(){return mv->Duration();};
  void SetId(double id) {stream_id = id;};
  void CloneStats(Virtualstream *);
  Virtualstream *GenerateBogus();
  void SwitchId(Virtualstream *v) 
    { double tmpid = stream_id; stream_id = v->StreamId(); v->SetId(tmpid);};
  void SetInitialWaitTime(double);
  void UpdateWaitlistStat(WaitStat *);
  void UpdateQuitters(WaitStat *, Subscriberlist *, Subscriberlist *);

  Virtualstream(Channel *c, Movie *m);
  virtual ~Virtualstream();

  Status Getstatus(void) {return status;};
  void Setstatus(Status s) {status = s;};
  void Setuntil(double u) {until_time =u;};
  void Adduser(Subptr *s) 
    {s->sptr->SetMystream(this); users.Prepend(s , NULL);};
  Subptr *Detachuser(Subptr *s) 
    {s->sptr->SetMystream(NULL); return((Subptr *)users.Detach(s));};
  Subptr *Findsubptr(Subscriber *u);
  Virtualstream *Forkstream(Channel *c, Status s, double u);
  void Joinstream(Virtualstream *v, VSList *, ChannelManager *);
  int Matchstream(Virtualstream *v);
  int NumUsers(void) {return users.Numnodes();};
  double GetPos(void) { return position;};
  double StreamId(void) {return stream_id;};
  void Advancestream(double timenow, double adv);
  virtual void Dumpdata(void);

  void SetPos(double pos) {position = pos;};
  Cost GetCostBandwidth();
  void CheckVideoTime(double);
  
};

class VSList: public List
{
public:
  static int num_vslist;

  // total Time to Recomputation; and number of Streams for Time to Recomputation
  // auxiliary variables used to compute how much time on the average a stream
  // waits until a new recomputation interval starts.

  double totalTTR;
  double nStreamsTTR;

  VSList() {num_vslist++; totalTTR = nStreamsTTR = 0.0;};
  ~VSList() {num_vslist--;};
  void Advanceall(double curtim);
  void Launchbatch(VSList *b, double curtim);
  void Handleexits(Subscriberlist *active, Subscriberlist *idle, ChannelManager *cmgr);
  Virtualstream *GetMatch(Virtualstream *v);
  void Compact(ChannelManager *);
  void Aggregate(int policy, char recomp, double *window, double curtim);
  void GreedyMerge(STREAM *, int, double *);
  void ChangeStatus(double, Status);
  void ChangeInList(STREAM *, int, NStatus);
  Status GetStat(double);
  double GetPosition(double);
  int Present(double);

  int GetNumberVS();
  Virtualstream *GetVS(double);
  int GetNBatched();
  Cost listcost;
  void GetCostBandwidth();
  void ResetAdTime();
  void CheckReset();
  Virtualstream *Findsubptrvs(Subscriber *s);
  void RemoveBogus();
  void UsersPerStream();
};


// The ad stream list is needed because this list should NOT take the streams
// out of the rmvstream. This ad stream list should only be a reference to
// keep track which streams/users requested interaction while during ad time

class VSptr: public Node
{
  friend class AdStrList;
public:
  static int num_vsptr;
  Subscriber *subs;
  Status status;
  double duration;
  VSptr() {status=UNKNOWN; duration=0.0; subs=(Subscriber *)0; num_vsptr++;};
  VSptr(Subscriber *sb, Status st, double d) \
    {subs = sb; status = st; duration = d; num_vsptr++;};
  virtual ~VSptr() {num_vsptr--;};
  virtual void Dumpdata();
};

class AdStrList: public List
{
public:
  static int num_adstrlist;
  AdStrList() {num_adstrlist++;};
  ~AdStrList() {DeleteAllNodes();num_adstrlist--;};
  void Store(Subscriber *, Status, double);
  void Dispatch(ChannelManager *, Subscriberlist *,VSList *,double);
};

// For tracking stream bandwidth consumption
// Currently not used

class MStream: public Node
{
  friend class List;

public:
  double stream_id;
  double starting_position;
  double starting_time;
  double end_position;
  double end_time;
  double amount_in_play;
  double amount_in_ad;
  static int  num_mstream;

  MStream *getNext() { return (MStream *)next;};
  MStream *getPrev() { return (MStream *)prev;};

  MStream() {amount_in_play = amount_in_ad = 0.0;num_mstream++;};
  virtual ~MStream() {num_mstream--;};
};
  
#include "channel.hpp"

#endif //__VIRTSTREAM_HPP__
