/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu,
 * Wang Ke
 */

#include "virtstream.hpp"
#include "adins.hpp"
#include "channel.hpp"
#include "config.hpp"
#include <math.h>

extern int verbose; // debug information for Dumpdata - declared in main.cpp
extern double k;    // global simulation time - declared in main.cpp
extern Config config; // config variable containing configuration data - declared in main.cpp

void printChannelType(ChannelType which)
{
  switch(which) {
  case OnDem: 
    cout << "ON-DEMAND";
    break;
  case Cont:
    cout << "CONTINGENCY";
    break;
  case Both:
    cout << "BOTH";
    break;
  default:
    cout << "UNKNOWN";
    break;
  }
}

void Channel::Dumpdata()
{
  if (verbose) {
    cout << "Channel ";
    printChannelType(channelType);
    cout << " - " << channelID << " carrying virtualstream: ";

    if (v) v->Dumpdata();
    else cout << " NULL " << endl;
  }
}


// obtaining the next time an ad is scheduled to start
double Ad_Channel::get_nearest_ad_time(double timenow)
{
  double up;

  // if the policy is not ADINSPOL, then it doesn't matter
  // when the "ad_time" is
  if ( (config.data.AGG_POLICY != ADINSPOL) )
    return timenow;

  if ( timenow < last_ad_start_time ) return -1.0;
  
  up = last_ad_start_time + 
    ( (ceil(((timenow - last_ad_start_time)/ad_duration))) *ad_duration);

  // in the ad_channel, the streams have to wait until the next ad will start
  // that's why it is returning the "up" (using function ceil) ad time.

  return up;
}

Channellist::Channellist()
{
  num_channellist++;
}

Channellist::~Channellist()
{
  if (Numnodes()) DeleteAllNodes();
  num_channellist--;
}

void Channellist::buildList(int n)
{
  while(n > 0) {
    Prepend(new Channel(), NULL);
    n--;
  }
}

Channel *Channellist::detachChannel(int which)
{
  if (!num_nodes) return (NULL);
  Channel *cursor = (Channel *)head;
  while (cursor != (Channel *)0) {
    if (cursor->getID() == which) {
      return ( (Channel *) Detach(cursor));
    }
    cursor = cursor->getNext();
  }
  return (cursor);
}

Channel *Channellist::detachChannel()
{
  if (!num_nodes) return (NULL);
  return ( (Channel *)Detach(head));
}


ChannelManager::ChannelManager(int numchan, double frac)
{
  num_channelmanager++;
  fractionOnDemand = frac;
  totNumber = numchan;
  
  numOnDemand = (int)rint(frac*numchan);
  numContingency = numchan - numOnDemand;

  if (numContingency <= 0) {
    cerr << "ChannelManager: ERROR: Number of Contingency channels == "
	 << numContingency << endl;
    assert (numContingency > 0);
  }
  
  activeOnDemandlist.buildList(0);
  activeContingencylist.buildList(0);
  idlelist.buildList(totNumber);

  qmgr = new QueueManager();
  assert(qmgr);
}

ChannelManager::~ChannelManager()
{
  delete qmgr;
  num_channelmanager--;
}

int ChannelManager::onDemandAvailable()
{
  int n = activeOnDemandlist.Numnodes();
  assert((n >= 0));

  n = (numOnDemand > n) ? (numOnDemand - n) : 0;
  n = min(n, idlelist.Numnodes());
  assert( (n >= 0));
  return (n);
}

Channel *ChannelManager::getIdleChannel(ChannelType ctyp)
{
  switch(ctyp) {
  case OnDem:    
    if (onDemandAvailable() == 0) 
      return NULL;
    break;
  case Cont:
    if (contingencyAvailable() == 0) 
      return NULL;
    break;
  default:
    cerr << "ERROR: ChannelManager::getIdleChannel - wrong type\n";
    return NULL;
  }
  
  Channel *c = idlelist.detachChannel();
  c->setType(ctyp);
  assert(c->assignedAt < 0.0);
  c->assignedAt = k;
  return c;
}

void ChannelManager::putToActive(Channel *ch)
{
  switch(ch->getType()) {
  case OnDem:
    activeOnDemandlist.Prepend(ch, NULL);
    break;
  case Cont:
    activeContingencylist.Prepend(ch,NULL);
    break;
  default:
    cout << "ERROR: ChannelManager::putToActive - wrong channel type ";
    printChannelType(ch->getType());
    assert( (ch->getType() == Cont) );
    break;
  }
}

void ChannelManager::returnActiveChannel(Channel *ch)
{
  Channel *c = NULL;

  assert(ch);
  switch (ch->getType()) {

  case OnDem:
    c = activeOnDemandlist.detachChannel(ch->getID());
    if (!c) cerr << "ChannelManager:: ERROR: returning a non-active OnDemand Channel\n";
    assert(c);
    break;
  case Cont:
    c = activeContingencylist.detachChannel(ch->getID());
    if (!c) cerr << "ChannelManager:: ERROR: returning a non-active Contingency Channel\n";
    assert(c);
    break;
  case Undef:
    cerr << "ChannelManager:: ERROR: returning an undefined type of channel\n";
    assert( (ch->getType() != Undef) );
    break;
  default:
    cerr << "ChannelManager:: ERROR: channel without proper channelType\n";
    ch->Dumpdata();
    assert( (ch->getType() == OnDem) );
    break;
  }

  c->setType(Undef);
  c->setVS(NULL);
  c->assignedAt = -1.0;

  idlelist.Prepend(c,NULL);
}
    
void ChannelManager::waitForOnDemandChannel(Virtualstream *vs)
{
  assert(vs);
  vs->SetInitialWaitTime(k);

  // no need to call  "joinVS" here because when serveOnDemandRequests
  // is called, it checks against the batched streams
  qmgr->onDemandlist.Prepend(vs,NULL); 
  qmgr->onDemandlist.wstat.stream_waiting++;
}

int ChannelManager::contingencyAvailable()
{
  int i = activeContingencylist.Numnodes();
  assert(i>=0);

  i = ( (numContingency > i) ? (numContingency - i) : 0);
  i = min(i, idlelist.Numnodes());
  assert( (i>=0));
  return (i);
}

void ChannelManager::waitForContingencyChannel(Virtualstream *vs)
{
  assert(vs);
  vs->SetInitialWaitTime(k);
  qmgr->contingencylist.Prepend(vs,NULL);
  qmgr->contingencylist.wstat.stream_waiting++;
}

void ChannelManager::changeFraction(double newfraction)
{
  assert(newfraction >= 0.0);
  assert(newfraction <= 1.0);

  fractionOnDemand = newfraction;
  numOnDemand = (int)rint(fractionOnDemand*totNumber);
  numContingency = totNumber - numOnDemand;

  assert( ((numContingency + numOnDemand) == totNumber ) );
}

void ChannelManager::serveOnDemandRequests(VSList *bmvstreams)
{
  int m = onDemandAvailable();
  int n = qmgr->onDemandlist.Numnodes();
  Channel *c;
  Virtualstream *vv, *ww;

  while ( (n > 0) && (m > 0) ) {

    c = getIdleChannel(OnDem);
    assert(c);
    vv = qmgr->onDemandlist.detachVirtualstream();
    assert(vv);
    
    vv->SetChan(c);
    c->setVS(vv);
    putToActive(c);

    n--;
    qmgr->onDemandlist.wstat.stream_waiting--;
    assert ( (qmgr->onDemandlist.wstat.stream_waiting >= 0) );

    if((ww=bmvstreams->GetMatch(vv))) {
      ww->Joinstream(vv, bmvstreams, this); 
      ww->n_batched += 1; 
      assert( (vv->GetChan() == 0));
      assert( (c->getType() == OnDem));
      
    } 
    else {
      bmvstreams->Prepend(vv, NULL); 
      m--;
    }
  }
  
  assert( ( (onDemandAvailable() == 0) || (qmgr->onDemandlist.Numnodes() == 0) ) );
}

void ChannelManager::serveContingencyRequests(VSList *mvstreams)
{
  int m = contingencyAvailable();
  int n = qmgr->contingencylist.Numnodes();
  Channel *c;
  Virtualstream *vv, *ww;

  while ( (n > 0) && (m > 0) ) {

    c = getIdleChannel(Cont);
    assert(c);
    vv = qmgr->contingencylist.detachVirtualstream();
    assert(vv);
    
    vv->SetChan(c);
    c->setVS(vv);
    putToActive(c);

    n--;
    qmgr->contingencylist.wstat.stream_waiting--;
    assert( (qmgr->contingencylist.wstat.stream_waiting >= 0) );

    if((ww=mvstreams->GetMatch(vv))) {
      ww->Joinstream(vv, mvstreams, this); 
      assert( (vv->GetChan() == 0));
      assert( (c->getType() == Cont));

    } 
    else {
      mvstreams->Prepend(vv, NULL); 
      m--;
    }
  }
  
  assert( ( (contingencyAvailable() == 0) || (qmgr->contingencylist.Numnodes() == 0) ) );
}

void ChannelManager::checkWaitUntilLimits(Subscriberlist *active, Subscriberlist *idle)
{
  qmgr->onDemandlist.compact(active, idle);
  qmgr->contingencylist.compact(active, idle);
}

void Waitlist::appendVirtualstream(Virtualstream *nvs)
{
  Prepend(nvs, NULL);
}

Virtualstream * Waitlist::detachVirtualstream()
{
  if (Numnodes() <= 0) return NULL;

  Virtualstream *vs = (Virtualstream *)head;
  vs->UpdateWaitlistStat(&wstat);

  return ((Virtualstream *) Detach(head));
}

double Waitlist::getAverageWT()
{
  if (wstat.total_num_wait == 0.0) 
    return -1.0;
  return ( (wstat.total_wait_time/wstat.total_num_wait));
}

int Waitlist::joinVS(Virtualstream *nvs, ChannelManager *cmgr)
{
  if (Numnodes() <= 0) return 0;

  Virtualstream *cursor = (Virtualstream *) Gethead();

  while (cursor) {
    if (cursor->Matchstream(nvs)){
      if ( (!cursor->bogus) && (!nvs->bogus) ) {
	cursor->Joinstream(nvs, 0, cmgr);
	return 1;
      }
    }
  }
  
  return 0;
}

void Waitlist::compact(Subscriberlist *active, Subscriberlist *idle)
{
  if (Numnodes() <=0) return;
  Virtualstream *cursor = (Virtualstream *) Gethead();
  Virtualstream *tmpcursor;
  
  while(cursor) {
    if (cursor->NumUsers() > 0) {
      cursor->UpdateQuitters(&wstat, active, idle);
      if (cursor->NumUsers() <= 0) {
	tmpcursor = cursor;
	cursor = (Virtualstream *) cursor->Getnext();
	
	assert( (tmpcursor->GetChan() == NULL) );
	Remove(tmpcursor);
	wstat.stream_waiting--;
	assert( (wstat.stream_waiting >= 0) );
      }	
      else 
	cursor = (Virtualstream *) cursor->Getnext();
    }
    else {
      cerr << "ERROR: Waitlist:compact: virtualstream with numusers <= 0\n";
      cursor->Dumpdata();
      assert( (cursor->NumUsers() > 0) );
    }
  }
  return;
}

double QueueManager::getAverageWT(ChannelType which) {

  switch(which) {

  case OnDem:
    return (onDemandlist.getAverageWT());
  case Cont:
    return (contingencylist.getAverageWT());
  case Both:
    if ( ((onDemandlist.wstat.total_num_wait + contingencylist.wstat.total_num_wait)
	  == 0.0) )
      return -1.0;
    return( (onDemandlist.wstat.total_wait_time + contingencylist.wstat.total_wait_time) /
	    (onDemandlist.wstat.total_num_wait + contingencylist.wstat.total_num_wait) );
  default:
    cerr << "ERROR: QueueManager::getAverageWT failed switch statement. ChannelType called: "
	 << (int)which << endl;
  }
  return 0.0;
}

double QueueManager::getPercUp(ChannelType which) {

  switch(which) {

  case OnDem:
    if (onDemandlist.wstat.total_num_wait == 0.0)
      return -1.0;
    return (onDemandlist.wstat.moreUp/onDemandlist.wstat.total_num_wait);
  case Cont:
    if (contingencylist.wstat.total_num_wait == 0.0)
      return -1.0;
    return (contingencylist.wstat.moreUp/contingencylist.wstat.total_num_wait);
  case Both:
    if ( ((onDemandlist.wstat.total_num_wait + contingencylist.wstat.total_num_wait) 
	 == 0.0) ) 
      return -1.0;
    return( (onDemandlist.wstat.moreUp + contingencylist.wstat.moreUp) /
	    (onDemandlist.wstat.total_num_wait + contingencylist.wstat.total_num_wait) );
  default:
    cerr << "ERROR: QueueManager::getPercUp failed switch statement. ChannelType called: "
	 << (int)which << endl;
  }
  return 0.0;
}


double QueueManager::getPercLow(ChannelType which)
{
  switch(which) {

  case OnDem:
    if (onDemandlist.wstat.total_num_wait == 0.0)
      return -1.0;
    return (onDemandlist.wstat.lessLow/onDemandlist.wstat.total_num_wait);
  case Cont:
    if (contingencylist.wstat.total_num_wait == 0.0)
      return -1.0;
    return (contingencylist.wstat.lessLow/contingencylist.wstat.total_num_wait);
  case Both:
    if ( (onDemandlist.wstat.total_num_wait + contingencylist.wstat.total_num_wait) 
	 == 0.0 ) 
      return -1.0;
    return( (onDemandlist.wstat.lessLow + contingencylist.wstat.lessLow) /
	    (onDemandlist.wstat.total_num_wait + contingencylist.wstat.total_num_wait) );
  default:
    cerr << "ERROR: QueueManager::getPercLow failed switch statement. ChannelType called: "
	 << (int)which << endl;
  }
  return 0.0;
}


double QueueManager::getUp(ChannelType which) 
{
  switch(which) {

  case OnDem:
    return (onDemandlist.wstat.moreUp);
  case Cont:
    return (contingencylist.wstat.moreUp);
  case Both:
    return( (onDemandlist.wstat.moreUp + contingencylist.wstat.moreUp));
  default:
    cerr << "ERROR: QueueManager::getUp failed switch statement. ChannelType called: "
	 << (int)which << endl;
  }
  return 0.0;
}

double QueueManager::getLow(ChannelType which)
{
  switch(which) {

  case OnDem:
    return (onDemandlist.wstat.lessLow);
  case Cont:
    return (contingencylist.wstat.lessLow);
  case Both:
    return( (onDemandlist.wstat.lessLow + contingencylist.wstat.lessLow));
  default:
    cerr << "ERROR: QueueManager::getLow failed switch statement. ChannelType called: "
	 << (int)which << endl;
  }
  return 0.0;
}

double QueueManager::getTotal(ChannelType which)
{
  switch(which) {

  case OnDem:
    return (onDemandlist.wstat.total_num_wait);
  case Cont:
    return (contingencylist.wstat.total_num_wait);
  case Both:
    return( (onDemandlist.wstat.total_num_wait + contingencylist.wstat.total_num_wait) );
  default:
    cerr << "ERROR: QueueManager::getTotal failed switch statement. ChannelType called: "
	 << (int)which << endl;
  }
  return 0.0;
}

double QueueManager::getQuit(ChannelType which)
{
  switch (which) {
  case OnDem:
    return (onDemandlist.wstat.tot_quitters);
  case Cont:
    return (contingencylist.wstat.tot_quitters);
  case Both:
    return (onDemandlist.wstat.tot_quitters + contingencylist.wstat.tot_quitters);
  default:
    cerr << "ERROR: QueueManager::getQuit failed switch statement. ChannelType Undef\n";
  }
  return 0.0;
}

double QueueManager::getQuitWT(ChannelType which)
{
  switch (which) {
  case OnDem:
    if ( (onDemandlist.wstat.tot_quitters == 0.0) ) 
      return -1.0;
    return ( (onDemandlist.wstat.tot_quit_wt/onDemandlist.wstat.tot_quitters) );
  case Cont:
    if ( (contingencylist.wstat.tot_quitters == 0.0) )
      return -1.0;
    return ( (contingencylist.wstat.tot_quit_wt/contingencylist.wstat.tot_quitters) );
  case Both:
    if ( ((onDemandlist.wstat.tot_quitters + contingencylist.wstat.tot_quitters)
	 == 0.0) )
      return -1.0;
    return ( ( (contingencylist.wstat.tot_quit_wt + onDemandlist.wstat.tot_quit_wt)/
	       (contingencylist.wstat.tot_quitters + onDemandlist.wstat.tot_quitters)));
  default:
    cerr << "ERROR: QueueManager::getQuitWT failed. ChannelType Undef\n";
  }
  return 0.0;
}

