/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <math.h>
#include "stream.h"
#include "afl.h"

extern CLUSTER *ef_cluster;
extern int     ef_chnl, ef_nclus;
extern int     ef_nstreams;
extern STREAM  *ef_stream;
extern double  bw_without_merge, bw_with_merge, total_bw_recl;
extern double  NetBandwidth;

extern double  AFL_cost;

/***************************************************************************/
/* calc_bandwidth_net calculates the net bandwidth used by a cluster       */
/***************************************************************************/
double calc_bandwidth_net(CLUSTER cluster, STREAM *streams, double trec)
{
    int i,j,k;
    double s;

    j = cluster.firstmemind;
    k = cluster.lastmemind;
    
    s = trec;
    for (i=j+1; i<=k; i++) {
	s += rint((Kfactor*(streams[j].position - streams[i].position))/0.0333);
    }
    return s;
}


/* For EMCL + AFL */
void call_afl(STREAM *streams, CLUSTER *clusters, double window)
{ 
    int i, j, first, last;
    int num_in_clust = 0;
    
    for (i=0; i<ef_nclus; i++) {
	first = clusters[i].firstmemind;
	last = clusters[i].lastmemind;
	streams[first].status = NORMAL;
	
	for (j=first+1; j<=last; j++)
	    streams[j].status = ACCEL;
    }
    
    /* Once the clusters are built, find out the net bandwidth reclaimed per cluster */
    for (i=0; i<ef_nclus; i++)    {
	int num = clusters[i].lastmemind - clusters[i].firstmemind + 1;
	num_in_clust += num;

#ifdef VERBOSE
	cout << "-----------------------------------------------------------" << endl;;
        cout << "Cluster number= "<< i << " Movie ID= " << streams[clusters[i].firstmemind].movie_id << endl
	     << " First stream= " << clusters[i].firstmemind
	     << " Program Position of 1st stream= " << streams[clusters[i].firstmemind].position << endl
	     << " Last stream= " << clusters[i].lastmemind
	     << " Program Position of last stream= " << streams[clusters[i].lastmemind].position << endl;
	print_stream_id(streams, clusters[i]);
#endif
	AFL_cost = rint(calc_bandwidth_net(clusters[i], streams, window));
	clusters[i].bandwidth_reclaimed = rint(num*window) - AFL_cost;
	total_bw_recl += clusters[i].bandwidth_reclaimed;
	NetBandwidth += AFL_cost;
	
#ifdef VERBOSE
	cout << "Cost of the AFL tree = " << AFL_cost << endl;
	cout << "Bandwidth reclaimed over a window of "<< window*0.0333 << " seconds = "  <<
	    clusters[i].bandwidth_reclaimed << " frames" << endl;
#endif
    }

    /* Add the additional bandwidth due to non clustered streams */
    NetBandwidth += rint((ef_nstreams - num_in_clust)*window);
}


/* Get Stream Status after a specified interval of time*/
STREAM  get_afl_state(double stream_id, STREAM *str, double time, CLUSTER *cluster, 
		       int nstreams)

{
    int i, index;
    double pos1, pos2;
    STREAM st;

    index = find_stream_index(stream_id, str, nstreams);

    if (index == -1) {
	cout << "This stream id not found in the table" << endl;
	exit(0);
    }

    st.movie_id = str[index].movie_id;

    if (!index) {
	st.status = NORMAL;
	st.position = str[index].position + time;
	return st;
    }
    
    for (i=0; i<(nstreams/2); i++)
	if ((index >= cluster[i].firstmemind) && (index <= cluster[i].lastmemind))
	    break;
    
    if (i==(nstreams/2)) { /* Not in any cluster */
	st.status = NORMAL;
	st.position = str[index].position + time;
	return st;
    }
    
    
    pos1 = str[cluster[i].firstmemind].position;
    pos2 = str[cluster[i].lastmemind].position;

    /* If all streams have merged OR first stream of the cluster */
    if ((Kfactor*(pos1 - pos2) <= time) || (index == cluster[i].firstmemind))
    {
	st.status = NORMAL;
	st.position = pos1 + time;
	return st;
    }
    
    /* If last stream of the cluster */
    if (index == cluster[i].lastmemind)
    {
	st.status = ACCEL;
	st.position = pos2 + ((Kfactor+1)*time)/Kfactor;
	return st;
    }
	
    /* General Case */
    if (Kfactor*(pos1 - str[index].position) <= time)
    {
	st.status = NORMAL;
	st.position = pos1 + time;
    } else {
	st.status = ACCEL;
	st.position = str[index].position + ((Kfactor+1)*time)/Kfactor;
    }
      
    return st;
}


/* Memory allocation function */
void ef_alloc()
{
    if((ef_stream = (STREAM *)calloc(ef_nstreams, sizeof(STREAM))) == NULL)
    {
	perror("Error allocating streams");
	exit(1);
    }
    
    if((ef_cluster = (CLUSTER *)calloc(ef_nstreams/2, sizeof(CLUSTER))) == NULL)
    {
	perror("Error allocating clusters");
	exit(1);
    }
}

/* Freeing memory */
void ef_free()
{
    if (ef_cluster)
    {
	free (ef_stream);
	free (ef_cluster);
    }
}

