/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu,
 * Wang Ke
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "stream.h"
#include "rsma.h"
#include "afl.h"

#include "adins.hpp"
#include "interface.h"
#include "config.hpp"

/* Following correspond to the EMCL+RSMA policy */ 
int     **er_rsma, **r_rsma;
CLUSTER *er_cluster, *ef_cluster, *r_cluster;
GRAPH   *er_graph, *r_graph;
int     er_chnl, er_nclus, ef_chnl, ef_nclus;
int     r_chnl, r_nclus;

int     er_nstreams, ef_nstreams, r_nstreams;
STREAM  *er_stream, *ef_stream, *r_stream;
double  TIMEBASE;
double  bw_without_merge=0, bw_with_merge=0, total_bw_recl=0;
double  NetBandwidth = 0;

double  RSMA_cost, AFL_cost;

extern Ad_Insertion global_ad_ins;
extern Config config;

/* Returns the status of a particular stream at a point in the simulation*/
NStatus get_stream_status(double stream_id, double time, double position, Policy policy)
{
    int i;
    STREAM st;

    switch  (policy)
    {
    case EMCLRSMA:
    case EMCLRSMAHEUR:
	if ((i = find_stream_index(stream_id, er_stream, er_nstreams)) < 0) {
	    output_streams(er_stream, er_nstreams);
	    cout << endl << "Could not find index " << stream_id << endl;
	    break;
	}
	
	st = get_rsma_state(stream_id, er_stream, time-TIMEBASE, 
			    er_cluster, er_nstreams, er_graph, er_rsma);
	
	if (fabs((st.position - position)/0.0333) > 1)
	    return INVALID;

	return (st.status);
	
    case EMCLAFL:
	if ((i = find_stream_index(stream_id, ef_stream, ef_nstreams)) < 0) {
	    output_streams(ef_stream, ef_nstreams);
	    cout << endl << "Could not find index " << stream_id << endl;
	    break;
	}

	st = get_afl_state(stream_id, ef_stream, time-TIMEBASE, 
			   ef_cluster, ef_nstreams);
	
	if (fabs((st.position - position)/0.0333) > 1)
	    return INVALID;

	return (st.status);

    case GREEDY:
	break;

    case RSMA:
    case RSMAHEUR:
      if ((i = find_stream_index(stream_id, r_stream, r_nstreams)) < 0) {
	output_streams(r_stream, r_nstreams);
	cout << endl << "Could not find index " << stream_id << endl;
	break;
      }
      
      st = get_rsma_state(stream_id, r_stream, time-TIMEBASE, 
			  r_cluster, r_nstreams, r_graph, r_rsma);
      
      if (fabs((st.position - position)/0.0333) > 1)
	return INVALID;
      
      return (st.status);
      
    case ADINSPOL:
      return (global_ad_ins.get_stream_status(stream_id, position, time));
      break;
    default: break;
    }
    
    return INVALID;
}


/* Basically a despatcher function */
void invoke_policy(Policy policy, STREAM *stream, int num_streams, 
		   double window, double curr_time)
{
	switch  (policy)
	{
	case EMCLRSMA:
		TIMEBASE = curr_time;
		er_nstreams = num_streams;
		er_free();
		er_alloc();
		memcpy(er_stream, stream, er_nstreams*sizeof(STREAM));
		cluster(er_stream, er_nstreams, er_cluster, window, &er_nclus, &er_chnl);	
		call_rsma(er_stream, er_cluster, er_graph, er_rsma, window/0.0333);
		break;
	case EMCLAFL:
		TIMEBASE = curr_time;
		ef_nstreams = num_streams;
		ef_free();
		ef_alloc();
		memcpy(ef_stream, stream, ef_nstreams*sizeof(STREAM));
		cluster(ef_stream, ef_nstreams, ef_cluster, window, &ef_nclus, &ef_chnl);
		call_afl(ef_stream, ef_cluster, window/0.0333);
		break;
	case EMCLRSMAHEUR:
		TIMEBASE = curr_time;
		er_nstreams = num_streams;
		er_free();
		er_alloc();
		memcpy(er_stream, stream, er_nstreams*sizeof(STREAM));
		cluster(er_stream, er_nstreams, er_cluster, window, &er_nclus, &er_chnl);	
		call_rsma_heur(er_stream, er_cluster, er_graph, er_rsma, window/0.0333);
		break;
	case GREEDY:
		break;
	case RSMA:
		TIMEBASE = curr_time;
		r_nstreams = num_streams;
		r_free();
		r_alloc();
		memcpy(r_stream, stream, r_nstreams*sizeof(STREAM));
		cluster_rsma(r_stream, r_nstreams, r_cluster, &r_nclus, &r_chnl);	
		call_rsma_only(r_stream, r_cluster, r_graph, r_rsma);
		if (Config::debug_predict)
		  output_streams(r_stream, r_nstreams);
		break;
	case RSMAHEUR:
		TIMEBASE = curr_time;
		r_nstreams = num_streams;
		r_free();
		r_alloc();
		memcpy(r_stream, stream, r_nstreams*sizeof(STREAM));
		cluster_rsma(r_stream, r_nstreams, r_cluster, &r_nclus, &r_chnl);	
		call_rsma_heur_only(r_stream, r_cluster, r_graph, r_rsma);
		break;
	case ADINSPOL:
		global_ad_ins.invoke_policy(stream, num_streams, window, curr_time);
		break;
	
	default: 
		break;
	}
}

