/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "stream.h"
#include "rsma.h"
#include "config.hpp"
#include "statfns.h"
#include "event.hpp"
#include "adins.hpp"

extern Config config;
extern MovieRequest mvrq;
extern double RSMA_cost;

extern STREAM *r_stream;
extern int r_nstreams;

extern double cum_hist;
extern unsigned int current;

extern double *simulation_time; // see main.cpp


/****************************************************************/
double expint(double x)
{
  double s1, s2, y;

  if (x <= 1.0)
    y = -log(x)+((((1.07857E-3*x-9.76004E-3)*x+5.519968E-2)*x-0.24991055)*x+0.99999193)*x-0.57721566;
  else
    {
      s1 = (((x+8.5733287401)*x+18.059016973)*x+8.6347608925)*x+0.2677737343;
      s2 = (((x+9.5733223454)*x+25.6329561486)*x+21.0996530827)*x+3.9584969228;
      y = (exp(-x)/x) * (s1/s2);
    }
  
  return y;
}

/****************************************************************/
/* Find the parent of two points in the RSA grid */
POINT parent(POINT child1, POINT child2)
{
  POINT ret;

  ret.x = (child1.x <= child2.x) ? child1.x : child2.x;
  ret.y = (child1.y <= child2.y) ? child1.y : child2.y;

  return ret;
}

/****************************************************************/
/* Creates the RSA points from the inter stream distances */
POINT *make_graph (CLUSTER cluster, STREAM *streams, int *length)
{
  int        i=0,j=0,k=0;
  double     total = 0;
  POINT      *graph;
  
  j = cluster.firstmemind;
  k = cluster.lastmemind;
  
  *length = k-j+1;
  graph = (POINT *) malloc(sizeof (POINT) * (k-j+1));
  
  total = rint((Kfactor*(streams[j].position - streams[k].position))/0.0333);

  for (i=j; i<=k; i++) {
    double xcoord = rint((Kfactor*(streams[i].position - streams[k].position))/0.0333);
    graph[i-j].x = xcoord;
    graph[i-j].y = total - xcoord;
  }    
  
  return &graph[0];
}

/****************************************************************/
void print_graph(POINT *graph, int length)
{
  int i;
  
  if (!graph) return;
  
  cout << "Graph Points are:" << endl;
  
  for (i=0; i<length; i++)
    cout << graph[i].x << "_" << graph[i].y << " ";
  
  cout << endl;
}


/****************************************************************/
int int_calculate(int i, int j, int n, int *array)
{
  return *(array + i*n+j-(i*(i+1))/2);
}

/****************************************************************/
void int_assign(int i, int j, int n, int *array, int val)
{
  *(array + i*n+j-(i*(i+1))/2) = val;
}

/****************************************************************/
double dbl_calculate(int i, int j, int n, double *array)
{
  return *(array + i*n+j-(i*(i+1))/2);
}

/****************************************************************/
void dbl_assign(int i, int j, int n, double *array, double val)
{
  *(array + i*n+j-(i*(i+1))/2) = val;
}

/****************************************************************/
void print_vector(int *arr, int len)
{
  int i;

  for (i=0; i<len; i++)
    printf("%d ", arr[i]);
  
  printf("\n");
}
/****************************************************************/

inline double predict_arrival()
{
  // Perfect prediction
  double tm = mvrq.getNextEventTime();
  cout << "PREDICT: next arrival " << tm << "\n";
  if (Config::debug_predict == 5)
    return (-Config::debug_predict_value);

  return (- (Kfactor+1)/Kfactor * (tm - (*simulation_time)));
  
}

/****************************************************************/
void calc_tree(int *opt_k, int *tree, int i, int j, int n, STREAM *stream, int first)
{
  int_assign(i, j, n, tree, 1);
  
  if (i==j) 
    int_assign(i, i, n, tree, 1);
  else if ((i+1)==j)
    {
      int_assign(i, j, n, tree, 1);
      stream[first+j].status = ACCEL;
    }
  else
    {
      int m = int_calculate(i, j, n, opt_k);
      calc_tree(opt_k, tree, i, m, n, stream, first);
      if (i != m) stream[first+m].status = ACCEL;
      
      calc_tree(opt_k, tree, m+1, j, n, stream, first);
      stream[first+j].status = ACCEL;
    }
}

/****************************************************************/
int argmin(int i, int j, int n, double *cost, int first, STREAM *stream)
{
  int k, retval=i;
  double opt = 0, old = pow(2,31)-1, term;
  
  
  if (Config::cost_formulation == BIT) {
    term = rint(max(stream[first+i].movie_length - 
		    (Kfactor*(stream[first+i].position - stream[first+j].position)),
		    0.0)/0.0333);
    
    for (k=i; k<j; k++){
      opt = dbl_calculate(i,k,n,cost) + dbl_calculate(k+1,j,n,cost) - term;
      
      if (opt < old){
	old = opt;
	retval = k;
      }
    }
  }
  else if (Config::cost_formulation == TIME_L_INF) {
    
    double lmbt = Config::cost_rate*Kfactor*(stream[first+i].position - stream[first+j].position);
    term = exp(-lmbt) - lmbt*expint(lmbt);
    
    for (k = i; k < j; k++) {
      
      opt = dbl_calculate(i,k,n,cost) + dbl_calculate(k+1,j,n,cost) - term;
      
      if (opt < old) {
	old = opt;
	retval = k;
      }
    }
  }
  else if (Config::cost_formulation == TIME_L_FIN) {
    
    term = exp(-Config::cost_rate*Kfactor*
	       (Config::cost_mvlen_multiplier*stream[first+i].movie_length - stream[first+i].position)/(Kfactor+1))
      - exp(-Config::cost_rate*Kfactor*(stream[first+i].position - stream[first+j].position));
    
    for (k = i; k < j; k++) {
      opt = dbl_calculate(i,k,n,cost) + dbl_calculate(k+1,j,n,cost) + term;
      
      if (opt < old) {
	old = opt;
	retval = k;
      }
    }
  }
  
  if (j>i) dbl_assign(i,j,n,cost,old);
  
  return retval;
}


/****************************************************************/
void compute_rsma(CLUSTER cluster, STREAM *stream, int **out_tree)
{
  int i, m, k=0;
  int *opt_k, *tree;
  double *cost;
  
  int first = cluster.firstmemind;
  int last = cluster.lastmemind;
  int n = last - first + 1;
  int num = (n*(n+1))/2;
  
  /* Allocation and Initialization Step */
  cost = (double *) calloc(num, sizeof(double));
  opt_k = (int *) calloc(num, sizeof(int));
  tree = (int *) calloc(num, sizeof(int));
  
  /* Dynamic Programming Algorithm */
  
  if (Config::cost_formulation == BIT) {
    
    double pij_cost = 0;

    for (i=0; i<n; i++){ 
      dbl_assign(i,i,n,cost, rint(max(stream[first+i].movie_length - stream[first+i].position,0.0)/0.0333));
    }

    for (i=0; i<n-1; i++) {
      pij_cost = Kfactor*(stream[first+i].position - stream[first+i+1].position);
      assert(pij_cost >= 0);
      pij_cost = rint(max(stream[first+i].movie_length - pij_cost,0.0)/0.0333);
      dbl_assign(i,i+1,n,cost, dbl_calculate(i,i,n,cost) + dbl_calculate(i+1,i+1,n,cost) - pij_cost);
    }

  }
  else if (Config::cost_formulation == TIME_L_INF) {
    
    dbl_assign(n-1, n-1, n, cost, 0.0);
    
    for (i=0;i<n-1;i++) {
      double lmbt = Config::cost_rate*Kfactor*(stream[first+i].position - stream[first+i+1].position);
      double term = exp(-lmbt) - lmbt*expint(lmbt);
      
      dbl_assign(i, i, n, cost, 0.0);
      dbl_assign(i, i+1, n, cost, 0.0 - term);
    }
  }
  else if (Config::cost_formulation == TIME_L_FIN) {
    for (i=0;i<n-1;i++) {
      dbl_assign(i,i,n,cost,
		 1 - exp(-Config::cost_rate*Kfactor*
			 (Config::cost_mvlen_multiplier*stream[first+i].movie_length - stream[first+i].position)/(Kfactor+1)));
      
      dbl_assign(i,i+1,n,cost, 
		 2 - exp(-Config::cost_rate*Kfactor*
			 (Config::cost_mvlen_multiplier*stream[first+i+1].movie_length - stream[first+i+1].position)/(Kfactor+1)) 
		 - exp(-Config::cost_rate*Kfactor*(stream[first+i].position - stream[first+i+1].position)));
    }
  }
  else {
    fprintf(stderr, "cluster/rsma.c: ERROR : undefined cost formulation: %d\n",
	    (int)Config::cost_formulation);
    exit(0);
  }    
  
  for (i = 0; i < n; i++) {
    int_assign(i,i,n,opt_k,i);
    int_assign(i,i,n,tree,1);
  }  
  
  for (m=2; m<n; m++) {
    for (i=0; i<n-m; i++) {
      /* calculate optimal costs */
      k = argmin(i, i+m, n, cost, first, stream);
      
      /* store the optimal value of k for every pair */
      int_assign(i, i+m, n, opt_k, k);
    }
  }	
  
  RSMA_cost = dbl_calculate(0,n-1,n,cost);
  
  calc_tree(opt_k, tree, 0, n-1, n, stream, first);
  
  *out_tree = tree;

  free(cost);
  free(opt_k);
}



/****************************************************************/
STREAM 
get_status(int id, int global_id, STREAM *stream, double time, GRAPH graph, int *tree)
{
  int i,j,ii,jj;
  STREAM st;
  
  if (graph.arr[0].x <= time/0.0333) {
    st.status = NORMAL;
    st.position = stream[global_id-id].position + time;
    
    return st; 
  }
  
  if (id==0)  {
    /* First stream in the cluster is always NORMAL */
    st.status = NORMAL;
    st.position = stream[global_id].position + time;
    return st;
  }
  else if (id==graph.num-1) {
    
    /* Last stream in the cluster is always FAST */
    st.status = ACCEL;
    st.position = stream[global_id].position + ((Kfactor+1)*time)/Kfactor;
    return st;
  }
  
  for (i=0; i<id; i++)
    if (int_calculate(i, id, graph.num, tree))
      break;
  
  if (i==id) {
    for (j=graph.num-1; j>id; j--)
      if (int_calculate(id, j, graph.num, tree))
	break;
  }
  
  if (i<id) { /* S_id is accelerating in the beginning */
    if ((graph.arr[i].x - graph.arr[id].x) >= time/0.0333) {
      st.status = ACCEL;
      st.position = stream[global_id].position + ((Kfactor+1)*time)/Kfactor;
      return st;
    }
    else {
      ii = i;
      
      for (; ;) {
	jj = graph.num - 1;
	while (!int_calculate(ii,jj,graph.num,tree)) jj--;
	if ((graph.arr[ii].x - graph.arr[jj].x) >= time/0.0333) {
	  st.status = NORMAL;
	  st.position = stream[global_id-id+ii].position + time;
	  return st;
	}
	else {
	  ii = 0;
	  while (!int_calculate(ii,jj,graph.num,tree)) ii++;
	  if ((graph.arr[ii].x - graph.arr[jj].x) >= time/0.0333) {
	    st.status = ACCEL;
	    st.position = stream[global_id-id+jj].position + 
	      ((Kfactor+1)*time)/Kfactor;
	    return st;
	  }
	  else continue; 
	}
      }
    }
    
  }
  else { /* S_id is normal status in the beginning */
    if ((graph.arr[id].x - graph.arr[j].x) >= time/0.0333) {
      st.status = NORMAL;
      st.position = stream[global_id].position + time;
      return st;
    }
    else {
      jj = j;
      
      for (; ;) {
	ii = 0;
	while (!int_calculate(ii,jj,graph.num,tree)) ii++;
	if ((graph.arr[ii].x - graph.arr[jj].x) >= time/0.0333) {
	  st.status = ACCEL;
	  st.position = stream[global_id-id+jj].position + 
	    ((Kfactor+1)*time)/Kfactor;
	  return st;
	}
	else {
	  jj = graph.num - 1;
	  while (!int_calculate(ii,jj,graph.num,tree)) jj--;
	  if ((graph.arr[ii].x - graph.arr[jj].x) >= time/0.0333) {
	    st.status = NORMAL;
	    st.position = stream[global_id-id+ii].position + time;
	    return st;
	  }
	  else continue; 
	}
      }
    }
  }
  
  return st; /* redundant statement: never reached */ 
}


/****************************************************************/
STREAM 
get_rsma_state(double stream_id, STREAM *str, double time, CLUSTER *cluster, 
	       int nstreams, GRAPH *graph, int **tree)
{
  int i, index;
  STREAM st;
  
  index = find_stream_index(stream_id, str, nstreams); 
  
  if (index == -1) {
    cout << "This stream id not found in the table" << endl;
    exit(0);
  }
  
  st.movie_id = str[index].movie_id;
  
  if (!index) {
    st.status = NORMAL;
    st.position = str[index].position + time;
    return st;
  }
  
  for (i=0; i<(nstreams/2); i++)
    if ((index >= cluster[i].firstmemind) && (index <= cluster[i].lastmemind))
      break;
  
  if (i==(nstreams/2)) { /* Not in any cluster */
    st.status = NORMAL;
    st.position = str[index].position + time;
    return st;
  }
  
  st = get_status(index - cluster[i].firstmemind, index, str, time, graph[i], tree[i]);
  st.movie_id = str[index].movie_id;
  
  return st;
}
/****************************************************************/
void cluster_rsma(STREAM *streams, int ns, CLUSTER *clusters, int *nclusters, int *nchannels)
{
  int i, j, nstreams = ns;
  
  if (nstreams == 1) 
    {
      *nchannels = 1;
      *nclusters = 0;
      return;
    }

  if (Config::predict)
    {
      /* WARNING: works for 1 movie only */
      streams = (STREAM *) realloc(streams, (nstreams+1)*sizeof(STREAM));
      r_stream = streams;
      r_nstreams = r_nstreams+1;

      streams[nstreams].movie_id = streams[0].movie_id;       
      streams[nstreams].stream_id = -1;
      streams[nstreams].position = predict_arrival();  
      streams[nstreams].status = ACCEL;
      streams[nstreams].movie_length = streams[0].movie_length;

      nstreams = ns + 1;
    }

  /* Sort streams -- can use counting sort, but I am lazier than thou */
  qsort(streams, nstreams, sizeof(STREAM), compar);
  
  /* Find all clusters and their count */
  for(i=1, j=0, *nclusters=0; i<nstreams; i++)
    {
      if(CARRY_SAME_PROGRAM(streams[i].movie_id, streams[j].movie_id))
	continue;
			
      if(j < i-1)
	{
	  clusters[*nclusters].firstmemind = j;
	  clusters[*nclusters].lastmemind = i-1;
	  (*nclusters)++;
	}
      j = i;
    }
  
  if(j < i-1)
    {
      clusters[*nclusters].firstmemind = j;
      clusters[*nclusters].lastmemind = i-1;
      (*nclusters)++;
    }
  
  if(!(*nclusters)) return;
  
  /* Sort clusters */
  qsort(clusters, *nclusters, sizeof(CLUSTER), comparclust);
  
  
  for(i=0, *nchannels=0; i<*nclusters; i++) 
    *nchannels += clusters[i].lastmemind - clusters[i].firstmemind; 
  
}


