/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Prithwish Basu
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include <unistd.h>

#include "stream.h"
#include "rsma.h"
#include "rsma_heur.h"

/*************************************************************************/
extern int     **er_rsma, **r_rsma;
extern CLUSTER *er_cluster, *r_cluster;
extern GRAPH   *er_graph, *r_graph;
extern int     er_chnl, er_nclus, r_chnl, r_nclus;
extern int     er_nstreams, r_nstreams;
extern STREAM  *er_stream, *r_stream;
extern double  TIMEBASE;
extern double  bw_without_merge, bw_with_merge, total_bw_recl;
extern double  NetBandwidth;

extern double  RSMA_cost;
struct timeval tv1, tv2;

/*************************************************************************/
/* For EMCL + RSMA */
void call_rsma(STREAM *streams, CLUSTER *clusters, GRAPH *graph, int **rsma, double W)
{
    int i, length;
    int num_in_clust = 0;

    for (i=0; i<er_nclus; i++)
    {
	graph[i].arr = make_graph(clusters[i], streams, &length);
	graph[i].num = length;
	num_in_clust += length;

	bw_without_merge = rint(length * W);

	gettimeofday(&tv1, (struct timezone *)0);
 	compute_rsma(clusters[i], streams, &rsma[i]);
	gettimeofday(&tv2, (struct timezone *)0);

	/* calculate reclaimed bandwidth */
	bw_with_merge = rint(RSMA_cost + W - (Kfactor*(streams[clusters[i].firstmemind].position
						       -streams[clusters[i].lastmemind].position))/0.0333);


	clusters[i].bandwidth_reclaimed =  bw_without_merge - bw_with_merge;

	total_bw_recl +=  bw_without_merge - bw_with_merge;
	NetBandwidth += bw_with_merge; /* Add the bandwidth due to the RSMA */
    }

    /* Add the additional bandwidth due to non clustered streams */
    NetBandwidth += rint((er_nstreams - num_in_clust)*W);
}




/* Calls RSMA on the whole stream cluster for every particular movie */
void call_rsma_only(STREAM *streams, CLUSTER *clusters, GRAPH *graph, int **rsma)
{
  int i, length;
  int num_in_clust = 0;

  for (i=0; i<r_nclus; i++)
    {
      graph[i].arr = make_graph(clusters[i], streams, &length);
      graph[i].num = length;
      num_in_clust += length;

      gettimeofday(&tv1, (struct timezone *)0);
      compute_rsma(clusters[i], streams, &rsma[i]);
      gettimeofday(&tv2, (struct timezone *)0);

      clusters[i].bandwidth_reclaimed =  bw_without_merge - bw_with_merge;

      total_bw_recl +=  bw_without_merge - bw_with_merge;
      NetBandwidth += bw_with_merge; /* Add the bandwidth due to the RSMA */
    }
}




/*************************************************************************/
/* For EMCL + RSMA Heuristic */
void call_rsma_heur(STREAM *streams, CLUSTER *clusters, GRAPH *graph, int **rsma, double W)
{
    int i, length;
    int num_in_clust = 0;

    for (i=0; i<er_nclus; i++)
    {
	graph[i].arr = make_graph(clusters[i], streams, &length);
	graph[i].num = length;
	num_in_clust += length;
	bw_without_merge = rint(length * W);

	gettimeofday(&tv1, (struct timezone *)0);
	compute_rsa_heur(graph[i].arr, length, &rsma[i]);
	gettimeofday(&tv2, (struct timezone *)0);

	/* calculate reclaimed bandwidth */
	bw_with_merge = rint(RSMA_cost + W - 
			     (Kfactor*(streams[clusters[i].firstmemind].position
				       -streams[clusters[i].lastmemind].position))/0.0333);


	clusters[i].bandwidth_reclaimed =  bw_without_merge - bw_with_merge;

	total_bw_recl +=  bw_without_merge - bw_with_merge;
	NetBandwidth += bw_with_merge;
    }

    /* Add the additional bandwidth due to non clustered streams */
    NetBandwidth += rint((er_nstreams - num_in_clust)*W);
}





/* Calls RSMA-heur on the whole stream cluster for every particular movie */
void call_rsma_heur_only(STREAM *streams, CLUSTER *clusters, GRAPH *graph, int **rsma)
{
    int i, length;
    int num_in_clust = 0;

    for (i=0; i<r_nclus; i++)
    {
	graph[i].arr = make_graph(clusters[i], streams, &length);
	graph[i].num = length;
	num_in_clust += length;

	gettimeofday(&tv1, (struct timezone *)0);
	compute_rsa_heur(graph[i].arr, length, &rsma[i]);
	gettimeofday(&tv2, (struct timezone *)0);

	clusters[i].bandwidth_reclaimed =  bw_without_merge - bw_with_merge;

	total_bw_recl +=  bw_without_merge - bw_with_merge;
	NetBandwidth += bw_with_merge; /* Add the bandwidth due to the RSMA */
    }
}






/*************************************************************************/
/* Memory allocation function */
void er_alloc()
{
    if((er_stream = (STREAM *)calloc(er_nstreams, sizeof(STREAM))) == NULL)
    {
	perror("Error allocating streams");
	exit(1);
    }
    
    if((er_cluster = (CLUSTER *)calloc(er_nstreams/2, sizeof(CLUSTER))) == NULL)
    {
	perror("Error allocating clusters");
	exit(1);
    }
	
    if ((er_graph = (GRAPH *) calloc(er_nstreams/2, sizeof(GRAPH))) == NULL)
    {
	perror("Error allocating graph array");
	exit(1);
    }

    if ((er_rsma = (int **) calloc(er_nstreams/2, sizeof(int *))) == NULL)
    {
	perror("Error allocating tree array");
	exit(1);
    }
}


/*************************************************************************/
/* Freeing memory */
void er_free()
{
    int i;

    if (er_cluster != NULL)
    {
	for (i=0; i<er_nclus; i++)
	{
	    free (er_graph[i].arr);
	    free (er_rsma[i]);
	}
	
	free (er_stream); er_stream = NULL;
	free (er_cluster); er_cluster = NULL;
	free (er_graph); er_graph = NULL;
	free (er_rsma); er_rsma = NULL;
    }
}

/*************************************************************************/
void r_alloc()
{
    if((r_stream = (STREAM *)calloc(r_nstreams, sizeof(STREAM))) == NULL)
    {
	perror("Error allocating streams");
	exit(1);
    }
    
    if((r_cluster = (CLUSTER *)calloc(r_nstreams/2, sizeof(CLUSTER))) == NULL)
    {
	perror("Error allocating clusters");
	exit(1);
    }
	
    if ((r_graph = (GRAPH *) calloc(r_nstreams/2, sizeof(GRAPH))) == NULL)
    {
	perror("Error allocating graph array");
	exit(1);
    }

    if ((r_rsma = (int **) calloc(r_nstreams/2, sizeof(int *))) == NULL)
    {
	perror("Error allocating tree array");
	exit(1);
    }
}


/*************************************************************************/
/* Freeing memory */
void r_free()
{
    int i;

    if (r_cluster != NULL)
    {
	for (i=0; i<r_nclus; i++)
	{
	    free (r_graph[i].arr);
	    free (r_rsma[i]);
	}
	
	free (r_stream); r_stream = NULL;
	free (r_cluster); r_cluster = NULL;
	free (r_graph); r_graph = NULL;
	free (r_rsma); r_rsma = NULL;
    }
}

/*************************************************************************/






