/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan, 
 * Prithwish Basu,
 * Wang Ke
 */

#include "vod.hpp"

#include "config.hpp"
#include "adins.hpp"
#include "interface.h"
#include "virtstream.hpp"

extern double k;
extern Config config;

extern double *history;
extern double cum_hist;
extern unsigned int current;

extern void print_Stts(Status s);

static const double SHORT_INTERV = 1;
static const double LONG_INTERV = 20;
static double which_interv = SHORT_INTERV;

MovieRequest::MovieRequest()
{
  event_time = 0.0;
  rate = 0;
  ml = NULL;
  sl = NULL;
}

MovieRequest::MovieRequest(double r, Movielist *m, Subscriberlist *s) 
{
  event_time = 0.0; 
  rate = r; 
  ml = m;
  sl = s;

  GenerateEvent();
};

void MovieRequest::ResetParameters(double r, Movielist *m, Subscriberlist *s)
{
  event_time = 0.0;
  rate = r;
  ml = m;
  sl = s;
  GenerateEvent();
}

// Exponentially Distributed Inter-Event times
void MovieRequest::GenerateEvent(void) 
{
  double tmp = exp_rand(rate);

  if (Config::predict) {
    unsigned int currmod = current % Config::MAXHISTORY;
    double first = history[currmod];
    history[currmod] = tmp;
    cum_hist += tmp - first;
    current ++;
  }

  if (Config::debug_predict == 1) {
    tmp = which_interv;
    which_interv = (which_interv == SHORT_INTERV) ? LONG_INTERV : SHORT_INTERV;
  }

  if (tmp >= DBL_MAX) {
    cout << "WARNING: event.C - MovieRequest::GenerateEvent generated DBL_MAX interval for movie request\n";
    event_time = tmp;
  }
  else {
    event_time += tmp;
  }
  
  if(ml) mvchoice = ml->Randomchoice();
  if(sl) { 
    requester =  sl->ExtractRandom();
  }
}

void MovieRequest::Dumpdata(void) 
{
  cout << "Event Time = " << event_time;
  cout << "\tArrival Rate = " << rate;
  cout << "\tMovie Chosen : \n";
  if(mvchoice) 
    mvchoice->Dumpdata();
  else
    cout << "NULL\n";
  cout << "Requesting User :\n";
  if(requester) 
    requester->Dumpdata();
  else
    cout << "NULL\n";
}

VCRActionRequest::VCRActionRequest(double r, double mdur, Request rq, Subscriberlist *s) 
{
  meanduration = mdur; 
  event_time = 0.0; 
  rtype = rq; 
  rate = r; 
  sl = s;
  requester = NULL;

  GenerateEventTime();
  GenerateEventProcess();
};

// Exponentially Distributed Inter-Event times
void VCRActionRequest::GenerateEvent(void) 
{
  GenerateEventTime();
  GenerateEventProcess();

}

void VCRActionRequest::GenerateEventTime(void) 
{
  double tmp = exp_rand(rate);

  if (tmp >= DBL_MAX) {
    cout << "WARNING: event.C - VCRActionRequest::GenerateEventTime generated DBL_MAX interval\n";
    event_time = tmp;
  }
  else {
    event_time += tmp;
  }
}
void VCRActionRequest::GenerateEventProcess(void)
{
  if(!sl) return;

  // Persistently find a user who is playing and has no action pending
  int found = 0;
  for(int j = 0; j < 100; j++) 
  {

    requester = sl->RandomChoice();
    if(!requester) {
      return;
    }
    // Because there is a call to Passivate in the GENERATE_VCR_ACTIONS
    // the requester could have request of type PLAY but actually be in a 
    // stream which is in FF or RW. 
    if(requester->GetRequest() != PLAY) {
      continue;
    }

    if(requester->GetMystream() == NULL) {
      cout << "WARNING: a requester from activesublist that has null stream\n";
      requester->Dumpdata();
      assert(requester->GetMystream());
    }


    // The requester might be waiting for a channel to become available
    // and in that case he cannot execute VCR actions. 
    if( (requester->GetMystream() != NULL) ) {
      if (requester->GetMystream()->GetChan() == NULL) {
	continue;
      }
    }

    // if stream is in PLAYING or FASTPLAY or AD_INS then the requester can
    // be used
    if((requester->GetMystream()->Getstatus() != PLAYING) 
       && (requester->GetMystream()->Getstatus() != FASTPLAY)
       && (requester->GetMystream()->Getstatus() != AD_INS)
       ) {
      continue;
    }
    found = 1;
    break;
  }
  // Set type and duration of interaction

  if (!found) {
    requester = NULL;
    return;
  }

  requester->SetRequest(rtype);

  if (requester->GetMystream()->GetChan() == NULL) {
    cout << "How come?\n";
  }

  double tmp;
  tmp = exp_rand(1.0/meanduration);

  if (tmp >= DBL_MAX) {
    cout << "WARNING: VCRActionRequest::GenerateEventProcess generated a request with DBL_MAX duration\n";
  }

  if ( (config.data.AGG_POLICY == ADINSPOL) ) {
    switch(rtype) {
    case QUIT:
      break;
    case FFWD:
      // Ad Insertion:
      // requirement of tmp for FF: (FF_RATE-1)*tmp == m*AMIN, m integer
      // since FF_RATE == 5, if we allow m to be any integer, tmp might
      // be 7.5, which to be prevented, should put (FF..-1)*tmp = m*2*AMIN
      tmp =  rint((config.data.FF_RATE-1)*tmp/(2*AMIN));
      if (tmp == 0.0) tmp = 1.0;
      tmp = (tmp*(2*AMIN))/(config.data.FF_RATE-1);
      break;
    case RWIND:
      // Ad Insertion:
      // requirement of tmp for RW: (RW_RATE+1)*tmp == m*AMIN, m integer
      tmp = (rint((config.data.RW_RATE+1)*tmp/AMIN));    
      if (tmp == 0.0) tmp = 1.0;
      tmp = (tmp*AMIN)/(config.data.RW_RATE+1);
      break;
    case PAUSE:
      // Ad Insertion:
      // requirement of tmp for PAUSE, tmp == AMIN
      tmp = (rint(tmp/AMIN))*AMIN;
      if (tmp == 0.0) tmp = AMIN;
      break;
    default:
      break;
    }
  }
  requester->SetDuration(tmp);

}

void VCRActionRequest::pType(Request rqt)
{
  switch(rqt) {
  case (PLAY): cout << "PLAY"; break;
  case (QUIT): cout << "QUIT"; break;
  case (FFWD): cout << "FFWD"; break;
  case (RWIND): cout << "RWIND"; break;
  case (PAUSE): cout << "PAUSE"; break;
  case (UNKWN): cout << "UNKWN"; break;
  default: cout << "(not translated)"; break;
  }
}

void VCRActionRequest::pType()
{
  pType(rtype);
}

void VCRActionRequest::Dumpdata(void) 
{
  cout << "Event Time = " << event_time;
  cout << " Request Type = ";
  pType();
  cout << " (" << (int)rtype << ") " <<endl;
  cout << " Arrival Rate = " << rate;
  cout << " Mean duration = " << meanduration;
  cout << "\nRequesting User :\n";
  if(requester) 
    requester->Dumpdata();
  else
    cout << "NULL\n";
  cout << endl; 
}
