/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan, 
 * Prithwish Basu,
 * Wang Ke
 */

#include <iostream.h>
#include <assert.h>
#include "list.hpp"

extern int verbose;

Node *List::Detach(Node *nodptr)
{
  // if NULL do nothing
  if(!nodptr) {
    cerr << "WARNING: detach called on null node\n";
    return(NULL);
  }
  
  // Does this node belong to this list ??
  if(nodptr->mylist != this) {
    cerr << "WARNING: detach called on list that does not contain the node\n";
    if (verbose) {
      cout << "Node " << nodptr << " registered list " << nodptr->mylist 
	   << " : current list : " << this << endl;
      cout << "Checking node's information:\n";
      if (nodptr) nodptr->Dumpdata();
      cout << "Checking node's list's information:\n";
      if (nodptr->mylist) nodptr->mylist->Dumpdata();
      cout << "Checking list's information:\n";
      Dumpdata();
    }
    return(NULL);
  }

  nodptr->mylist = NULL;

  Node *fptr = nodptr->next;
  Node *bptr = nodptr->prev;

  nodptr->next = nodptr->prev = NULL;

  if(nodptr == head) head = fptr;
  if(nodptr == tail) tail = bptr;

  if(fptr) fptr->prev = bptr;
  if(bptr) bptr->next = fptr;

  num_nodes--;

  if (tail) {
    if ( (tail->next == tail) || (tail->prev == tail)) {
      cout << "ERROR: list.C: Detach - non-null tail pointing to itself\n";
      assert( (!tail));
    }
    assert( (tail->next == NULL) );
  }

  if (head) {
    if ( (head->prev == head) || (head->next == head)) {
      cout << "ERROR: list.C: Detach - non-null head pointing to itself\n";
      assert( (!head));
    }
    assert( (head->prev == NULL) );
  }
  return(nodptr);
}


// this separation is because some nodes - sptr, vsptr
// have pointers to nodes that should not be deleted

List::~List()
{
  if (!ptrlist) DeleteAllNodes();
  tot_num_lists--;
}

void List::DeleteAllNodes()
{
  if(!num_nodes) return;
  Node *cur = head;
  while(cur)
  {
    cur = head->next;
    delete Detach(head);
    head = cur;
    if (num_nodes < 0) {
      cout << "ERROR: list deleting more nodes than it has!\n\n";
      assert(num_nodes > 0);
    }
  } 
}

int List::Swap(Node *node1, Node *node2)
{
  if((node1 == NULL) || (node2 == NULL) || (node1 == node2))
    return(FAILURE);

  Node *temp = node1->next;

  Detach(node1);
  Prepend(node1, node2);
  Detach(node2);
  Prepend(node2, temp);

  return(SUCCESS);
}

Node *List::Seek(Node *datum, int steps, char direction)
{
  if(steps >= num_nodes) return(NULL);

  if(direction == FORWARD)
  {
    if(datum == NULL) datum = head;
    while(steps--)
    {
      if(!datum) return(NULL);
      datum = datum->next;
    }
  }
  else
  {
    if(datum == NULL) datum = tail;
    while(steps--)
    {
      if(!datum) return(NULL);
      datum = datum->prev;
    }
  }
  return(datum);
};

void List::Prepend(Node *nodptr, Node *before)
{
  if(!nodptr) return;
  nodptr->mylist = this;
  if(!num_nodes) {head = tail = nodptr; num_nodes++; return;}
  if(!before) {Append(nodptr, tail); return;}

  Node *bptr = before->prev;
  nodptr->next = before;
  nodptr->prev = bptr;
  before->prev = nodptr;
  if(bptr) bptr->next = nodptr;
  if(before == head) head = nodptr;

  if (nodptr->next == nodptr) {
    cout << "ERROR: list: nodptr->next == nodptr\n";
    assert(nodptr->next != nodptr);
  }
  if (nodptr->prev == nodptr) {
    cout << "ERROR: list: nodptr->prev == nodptr\n";
    assert(nodptr->prev != nodptr);
  }

  num_nodes++;
  assert( (tail->next == NULL) );
  assert( (head->prev == NULL) );
}

void List::Append(Node *nodptr, Node *after)
{
  if(!nodptr) return;

  assert( (nodptr != after) );

  nodptr->mylist = this;
  if(!num_nodes) {head = tail = nodptr; num_nodes++; return;}
  if(!after) {Prepend(nodptr, head); return;}

  Node *fptr = after->next;
  nodptr->next = fptr;
  nodptr->prev = after;
  after->next = nodptr;

  if(fptr) fptr->prev = nodptr;
  if(after == tail) tail = nodptr;

  if (nodptr->next == nodptr) {
    cout << "ERROR: list: nodptr->next == nodptr\n";
    assert(nodptr);
    assert(nodptr->next != nodptr);
  }
  if (nodptr->prev == nodptr) {
    cout << "ERROR: list: nodptr->prev == nodptr\n";
    assert(nodptr);
    assert(nodptr->prev != nodptr);
  }

  num_nodes++;
  assert( (tail->next == NULL) );
  assert( (head->prev == NULL) );
}

void List::Catenate(List *lst)
{
  if((!lst) || (lst == this)) return;
  if(!lst->num_nodes)
  {
    // delete lst;
    return;
  }

  Node *cursor = lst->head;

  if (num_nodes == 0) {
    head = tail = cursor;
    cursor->mylist = this;
    if (lst->num_nodes == 1) {
      num_nodes++;
      lst->num_nodes=0;
      lst->tail = lst->head = NULL;
      return;
    }
    cursor = cursor->next;
  }

  num_nodes += lst->num_nodes;
  lst->num_nodes = 0;

  lst->tail = lst->head = NULL;

  tail->next = cursor;
  cursor->prev = tail;

  while(cursor)
  {
    tail = cursor;
    cursor->mylist = this;
    cursor = cursor->next;
  }
  //  delete lst;
  return;
}

void List::Dumpdata(void)
{
  Node *cursor = head;
  cout << num_nodes << " Nodes\n";
  while(cursor)
  {
    cursor->Dumpdata();
    cursor = cursor->next;
  } 
}


