/* -*- Mode:C++ -*- */
/*
 * Copyright (c) Multimedia Communications Lab, Boston University
 *
 * Simulation Program to Evaluate Service Aggregation in VOD
 * =========================================================
 *
 * This work was supported by NSF under grant No. NCR-9523958
 * 
 * DISCLAIMER:
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * MULTIMEDIA COMMUNICATIONS LAB, BOSTON UNIVERSITY OR THE CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *
 * Rajesh Krishnan, 
 * Prithwish Basu,
 * Wang Ke
 */

#include "stream.h"
#include "vod.hpp"
#include "adins.hpp"
#include "config.hpp"
#include "channel.hpp"

#include <assert.h>
#include <math.h>

extern void printChannelType(ChannelType which);
extern void print_NStts(NStatus ns);
extern void print_Stts(Status s);
extern void print_Stts(Status s, int i);
extern double IN_PLAYING_STT;
extern Ad_Insertion global_ad_ins;
extern Config config;

extern double k; // current time - declared in main.cpp;
extern double j; // simulation tick - declared in main.cpp;
extern int verbose; // declared in main.cpp
extern int newEvent; // keep track of newEvents for each simulation tick - declared in main.cpp


// debugging purposes
void VSptr::Dumpdata()
{
  cout << "VSPTR: DUMPING DATA:\nSubscriber:\n";
  if (subs) subs->Dumpdata();
  cout << "Status:\n";
  print_Stts(status);
  cout << "Duration:\n" << duration << "\n\n";
}

// When an event is generated the subscriber is stored by AdStrList
// and its action is delayed until the end of the current "ad grid time"
// (recall that events should be allowed to take place only during intervals
// of 30s, so that all events will still fall in the same event grid)
void AdStrList::Store(Subscriber *sptr, Status newstatus, double duration)
{
  if (sptr == (Subscriber *)0) {
    cerr << "AdStrList::Append - error: sptr null\n";
    return;
  }

  VSptr *v = new VSptr(sptr, newstatus, duration);
  
  Append(v, tail);
  assert((v->mylist == this));
}

// run the stored events generated in the previous 30 seconds
void AdStrList::Dispatch(ChannelManager *cmgr, Subscriberlist *sl, 
			 VSList *vsl, double curtim)
{
  if (num_nodes == 0) 
    return;

  VSptr *cursor = (VSptr *)head;
  VSptr *cindex = cursor;
  Virtualstream *ovs = NULL;
  while (cindex) {
    if (cursor->subs==NULL) {
      cout << "ERROR: subs equal to NULL\n";
      assert(0);
    }

    ovs = cursor->subs->GetMystream();

    // if the stream is not in AD_INS
    if ((ovs->Getstatus() != AD_INS)) {
      if ( (ovs->Findsubptr(cursor->subs))==NULL) {
	cout << "Error: subscriber not found in stream\n";
	cout << "Subscriber information:\n";
	cursor->subs->Dumpdata();
	cout << "Subscriber stream information:\n";
	cursor->subs->GetMystream()->Dumpdata();
	Virtualstream *nvs = vsl->Findsubptrvs(cursor->subs);

	if (nvs != NULL) {
	  cout << "Subscriber found in stream:\n";
	  nvs->Dumpdata();
	  global_ad_ins.debug_stream_id(nvs->StreamId());
	}
	else {
	  cout << "Subscriber not found in list:\n";
	}
	global_ad_ins.debug_stream_id(ovs->StreamId());
	assert(0);
      }

      sl->Append(cursor->subs, NULL);

      // if the stream has only 1 user and it has not suffered interaction
      // after the previous recomputation, it needs to create a "bogus" 
      // stream to carry the original stream's instructions before interaction
      // was requested. If this "bogus" stream is not created, other streams
      // that might have merged with this stream will be confused when the
      // merge does not take place.

      if (ovs->NumUsers() == 1) {
	if ( (config.data.AGG_POLICY == ADINSPOL) ){
	  if (ovs->been_interaction == 0) {
	    Virtualstream *nbvs = ovs->GenerateBogus();
	    nbvs->SwitchId(ovs);
	    nbvs->to_merge_with = global_ad_ins.to_merge_with(nbvs->StreamId());
	    vsl->Append(nbvs,NULL);
	    nbvs->SetChan(NULL);
	  }
	  ovs->been_interaction = 1;
	}
	ovs->Setstatus(cursor->status);
	// duration -1 because if a request is serviced at time 30 with 
	// duration of 5, then it should keep the state until time 34,
	// after which it goes back to PLAYING
	ovs->Setuntil(curtim + (cursor->duration-1));

	cmgr->returnActiveChannel(ovs->GetChan());
	ovs->SetChan(NULL);
	if (cmgr->contingencyAvailable()) {
	  Channel *c = cmgr->getIdleChannel(Cont);
	  if (!c) { 
	    cerr << "AdStrList: Dispatch: ERROR: contingency available but "
		 << "pointer NULL\n";
	    assert(c);
	  }
	  ovs->SetChan(c);
	  cmgr->putToActive(c);
	}
	else {
	  cmgr->waitForContingencyChannel(ovs);
	}
      }
      else {
	Virtualstream *v = ovs->Forkstream(0, \
					   cursor->status,(curtim+(cursor->duration-1)));
	v->Adduser(ovs->Detachuser(ovs->Findsubptr(cursor->subs)));
	if ( (config.data.AGG_POLICY == ADINSPOL) )
	  v->been_interaction = 1;
	
	if (cmgr->contingencyAvailable()) {
	  Channel *c2 = cmgr->getIdleChannel(Cont);
	  if (!c2) {
	    cerr << "AdStrList: Dispatch: ERROR: contingency available but "
		 << "pointer NULL\n";
	    assert(c2);
	  }
	  ovs->SetChan(c2);
	  cmgr->putToActive(c2);
	  vsl->Prepend(v,NULL);
	}
	else {
	  cmgr->waitForContingencyChannel(ovs);
	}
      }
      
      cindex = (VSptr *)cursor->Getnext();
      Remove(cursor);
    }
    else { // the stream status is in AD_INS. The event is postponed until later
      cindex = (VSptr *)cursor->Getnext();
    }
    cursor = cindex;
  }
}  


// Initial values of strms and numstrms. "numstrms" is initialized with 0 so
// that if Aggregate is called with recomp = 0, numstrms will have a definite
// known value.
static STREAM *strms = NULL;
int numstrms = 0;

Virtualstream::Virtualstream(Channel *c, Movie *m)
{
  static double id = 0.0;
  id += 1.0;
  stream_id = id;
  chan = c;
  if (chan) chan->setVS(this);

  mv = m;
  position = 0.;
  status = INBATCH;
  until_time = 0;

  mylist = NULL;
  ad_to_go = Config::max_ad_time;
  time_to_go = Config::long_term_ad_time;
  s_video_to_go = IN_PLAYING_STT;
  video_to_go = IN_PLAYING_STT;
  n_ad_time = (int)Config::ratio_tprime_amax;
  to_merge_with = -1.0;
  continuous_video_time = -1.0;
  after_interaction = 0;
  been_interaction = 0;
  newArrival = 0;
  first_merge = 0;
  

  mstr = new MStream();
  mstr->stream_id = stream_id;
  mstr->starting_position = 0.0;
  mstr->starting_time = k;
  mstrlist = new List();

  mstrlist->ptrlist = 1;
  n_batched = 0;
  bogus = 0;
  num_vs++;

}

Virtualstream::~Virtualstream()
{
  if (stream_id == Config::stream_id) {
    cout << "Stream " << stream_id << " deleted at " << k << endl;
    Dumpdata();
  }
  if (stream_id == Config::snd_stream) {
    cout << "Stream " << stream_id << " deleted at " << k << endl;
    Dumpdata();
  }
  delete mstrlist;

  if (chan) 
    chan->setVS(NULL);  
  num_vs--;
}


void Virtualstream::SetChan(Channel *c)
{
  chan = c;
  if (c) c->setVS(this);
}

void Virtualstream::CloneStats(Virtualstream *v)
{
  chan = v->GetChan();
  mv = v->GetMv();
  position = v->GetPos();
  status = v->Getstatus();
  until_time = v->GetUntil();
  after_interaction = v->after_interaction;
  been_interaction = v->been_interaction;
  ad_to_go = v->ad_to_go;
  time_to_go = v->time_to_go;
  video_to_go = v->video_to_go;
  s_video_to_go = v->s_video_to_go;
  continuous_video_time = v->continuous_video_time;
  n_ad_time = v->n_ad_time;
  to_merge_with = v->to_merge_with;
  newArrival = v->newArrival;
  first_merge = v->first_merge;
  last_recomp = v->last_recomp;
}

Virtualstream *Virtualstream::GenerateBogus() 
{
  Virtualstream *v = new Virtualstream(chan,mv);
  v->CloneStats(this);
  v->bogus = 1;
  v->been_interaction = 0;
  return v;
}

// Set the users' initial wait time. This is when the stream
// is waiting for a channel.

void Virtualstream::SetInitialWaitTime(double dd)
{
  Subptr *cursor = (Subptr *)users.Gethead();
 
  if (!cursor) {
    if (verbose) 
      cout << "Virtualstream::SetInitialWaitTime - no users\n\n";
    return;
  }
  assert( (dd>=0.0) );
  while (cursor) {
    cursor->sptr->ResetWaitStats();
    cursor->sptr->start_wait_time = dd;
    cursor->sptr->SetWaitUntil();
    cursor = (Subptr *)cursor->Getnext();
  }
  return;
}


// Update the waitlist's statistics upon detaching a virtualstream

void Virtualstream::UpdateWaitlistStat(WaitStat *ws)
{
  if (users.Numnodes() <= 0) return;
 
  ws->total_num_wait += users.Numnodes();

  Subptr *cursor = (Subptr *)users.Gethead();

  if (!cursor) {
    if (verbose) 
      cerr << "ERROR: Virtualstream::getTotalWaitTime - users.Numnodes() = "
	   << users.Numnodes() << " but head NULL\n\n";
    return;
  }

  double tmp = 0.0;
  while (cursor) {
    // k is the current simulation time
    tmp = k - cursor->sptr->start_wait_time;
    assert ( tmp >= 0.0);
    ws->total_wait_time += tmp;
    if (tmp >= ws->upThreshold) ws->moreUp++;
    if (tmp <= ws->lowThreshold) ws->lessLow++;

    cursor->sptr->end_wait_time = k;
    cursor->sptr->ResetWaitStats();

    cursor = (Subptr *)cursor->Getnext();
  }

  return;
}

// check to see if the subscriber's patience ended and is quitting

void Virtualstream::UpdateQuitters(WaitStat *ws, Subscriberlist *act, Subscriberlist *idle)
{
  if (NumUsers() <= 0) return;

  Subptr *cursor = (Subptr *) users.Gethead();
  Subptr *tmpsptr;

  while (cursor) {
    
    if ( k >= cursor->sptr->wait_until) {
      ws->tot_quitters++;

      if (cursor->sptr->start_wait_time < 0.0) {
	cout << "Start Wait Time less than zero: ";
	cursor->sptr->Dumpdata();
	Dumpdata();
      }
      assert(cursor->sptr->start_wait_time >= 0.0);
      assert( (k-cursor->sptr->start_wait_time) >= 0.0);

      ws->tot_quit_wt += (k - cursor->sptr->start_wait_time);
      cursor->sptr->end_wait_time = k; // should be useless

      tmpsptr = cursor;
      cursor = (Subptr *) cursor->Getnext();

      tmpsptr->sptr->Passivate();
      tmpsptr->sptr->ResetWaitStats();
      idle->Prepend(act->Detach(tmpsptr->sptr), NULL);
      users.Remove(tmpsptr);
    }
    else {
      cursor = (Subptr *) cursor->Getnext();
    }
  }
  return;
}


Virtualstream *Virtualstream::Forkstream(Channel *c, Status s, double u)
{
  Virtualstream *v = new Virtualstream(c, mv);
  v->position = position;
  v->status = s;
  v->until_time = u;

  v->continuous_video_time = 0.0;
  v->ad_to_go = ad_to_go;
  v->time_to_go = time_to_go;
  v->s_video_to_go = s_video_to_go;
  v->video_to_go = video_to_go;
  v->n_ad_time = n_ad_time;
  v->mstr->starting_position = position;
  v->mstr->starting_time = k;

  return v;
}

void Virtualstream::Joinstream(Virtualstream *v, VSList *vsl, ChannelManager *cmgr)
{
  int info = Config::join_info;
  
  if (v->bogus) {
    assert( (config.data.AGG_POLICY == ADINSPOL) );

    if (info) {
      cout << "JOIN: " << k << " " << stream_id 
	   << " (this) called join on bogus " << v->StreamId(); 
    }
    if ( ((v->to_merge_with == StreamId()) && (been_interaction == 0)) ||
	 ((to_merge_with == v->StreamId()) && (been_interaction == 0)) ) {
      if (v->Getstatus() == AD_INS) {
	if( status == AD_INS) {
	  cout << "JOIN: ERROR: both streams in AD_INS status. This:\n";
	  Dumpdata();
	  cout << "v->\n";
	  v->Dumpdata();
	  assert(status !=AD_INS);
	}

	ad_to_go = v->ad_to_go;
	time_to_go = v->time_to_go;
	s_video_to_go = v->s_video_to_go;
	video_to_go = v->video_to_go;
	n_ad_time = v->n_ad_time;
	continuous_video_time = v->continuous_video_time;
      }
      else { // *this is in AD_INS, it should "disappear"
	stream_id = v->StreamId();
      }
      if ((to_merge_with == v->StreamId())) {to_merge_with = v->to_merge_with;}
      if (info) { cout << " - bogus deleted. New stream information:\n"; Dumpdata(); }
      delete v;
      return;
    }
    else {
      if (info) { 
	cout << " - NOT DELETED: v->status = ";
	print_Stts(v->Getstatus());
      }
      // return the bogus stream back to system
      // putting it before the "head" of the list so that it won't be 
      // processed again
      if (vsl) vsl->Append(v,NULL);
      else cerr << "ERROR: Virtualstream:Joinstream:  VSList vsl null but needed!\n\n";
      return;
    }
  }
  
  if (bogus) {
    assert( (config.data.AGG_POLICY == ADINSPOL) );

    if (info) {
      cout << "JOIN: " << k << " bogus " << stream_id 
	   << " (this) called join on " << v->StreamId(); 
    }
    
    if ( ((to_merge_with == v->StreamId()) && (v->been_interaction == 0)) ||
	 ((v->to_merge_with == stream_id) && (v->been_interaction == 0))) {

      MStream *tmp_mstrb = v->mstr;
      v->mstr = mstr;
      mstr = tmp_mstrb;
      
      mstrlist->Catenate(v->mstrlist);
      mstrlist->Append(v->mstr,0);
      
      Subptr *uu = (Subptr *)v->users.Gethead();
      while(uu)
	{
	  uu->sptr->SetMystream(this);
	  uu = (Subptr *)uu->Getnext();
	}
      users.Catenate(&(v->users));
      bogus = 0;

      chan = v->GetChan();
      assert(chan);
      chan->setVS(this);

      if (v->Getstatus() == AD_INS) {
	ad_to_go = v->ad_to_go;
	time_to_go = v->time_to_go;
	s_video_to_go = v->s_video_to_go;
	video_to_go = v->video_to_go;
	n_ad_time = v->n_ad_time;
	continuous_video_time = v->continuous_video_time;
      }
      else {
	stream_id = v->StreamId();
      }

      if (to_merge_with == v->StreamId()) { to_merge_with = v->to_merge_with;}
      if (info) { 
	cout << " - bogus deleted\n"
	     << "JOIN: new information:\n"; 
	Dumpdata();
      }
      delete v;
      return;
    }
    else {
      if (info) { 
	cout << " - NOT DELETED: v->status = ";
	print_Stts(v->Getstatus());
      }
      if (vsl) vsl->Append(v,NULL);
      else cerr << "ERROR: Virtualstream:Joinstream: VSList vsl null but needed!!\n\n";
      return;
    }
  }
      


  // the 2 joining streams should NOT be in the same status, ideally one is
  // in AD_INS and the other is in PLAYING. But there are cases in which
  // both have the same status, when they are BATCHED. And another case
  // is when the stream which suffered interaction comes back
  // to PLAYING state and can be joined to any other stream in 
  // PLAYING state that matches.

  if ( (config.data.AGG_POLICY == ADINSPOL) ) {
    if (( (v->status == status) && ((status != INBATCH) && (status!=PLAYING)))) {
      cout << "ERROR in assumption. Streams " << stream_id
	   << " and " << v->stream_id << " on merge: ";
      if (v->after_interaction) 
	cout << "v->after_interaction == 1\n";
      if (after_interaction)
	cout << "after_interaction == 1\n";
      
      cout << "Debugging stream " << StreamId() << endl;
      Dumpdata();
      global_ad_ins.debug_stream_id(StreamId());
      
      cout << "Debugging stream " << v->StreamId() << endl;
      v->Dumpdata();
      global_ad_ins.debug_stream_id(v->StreamId());
      
      print_Stts(status);
      assert(!( (v->status == status) 
		&& 
		((status!=INBATCH)&&(status!=PLAYING)) )
	     );
    }
    else { // can go on
      // first make sure that the streams after interaction are "absorbed"
      if ((v->after_interaction) || (after_interaction) ) {
	if (after_interaction) {
	  if (info) 
	    cout << "JOIN: " << k << " stream " << StreamId() 
		 << " in after_interaction. Invoked join to " << v->stream_id << endl;
	  stream_id = v->stream_id;
	  status = v->status;
	  until_time = v->until_time;
	  
	  MStream *tmp_mstr = v->mstr;
	  v->mstr = mstr;
	  mstr = tmp_mstr;
	  
	  ad_to_go = v->ad_to_go;
	  time_to_go = v->time_to_go;
	  s_video_to_go = v->s_video_to_go;
	  video_to_go = v->video_to_go;
	  n_ad_time = v->n_ad_time;
	  continuous_video_time = v->continuous_video_time;
	  after_interaction = v->after_interaction;
	  been_interaction = v->been_interaction;
	  
	  to_merge_with = v->to_merge_with;
	  
	  // free channel
	  assert(chan);
	  assert(chan->getType() == Cont);
	  assert(v->GetChan());

	  // "this" is the stream after interaction - it is occupying
	  // a contingency channel. If the option is for returning
	  // OnDemand channels, then the channel held by "v" must be 
	  // released

	  if (Config::return_ondemand) {
	    cmgr->returnActiveChannel(v->GetChan());
	    if (first_merge || v->first_merge ) {
	      first_merge = 0; v->first_merge = 0;
	    }
	  }
	  else {
	    cmgr->returnActiveChannel(chan);
	    if (first_merge || v->first_merge) {
	      first_merge = 0; v->first_merge = 0;
	    }
	    chan = v->GetChan();
	  }
	  assert(chan);
	  chan->setVS(this);
	  v->SetChan(NULL);
	  assert( ( (chan->getType() == OnDem) || (chan->getType() == Cont) ) );

	}
	else { // v is the stream after interaction
	  if (Config::join_info)
	    cout << "JOIN: " << k << " stream " << v->stream_id 
		 << " in after_interaction. Being joined to " << stream_id << endl;
	  
	  assert(chan);
	  assert(v->GetChan());
	  assert( (v->GetChan()->getType() == Cont) );

	  if (Config::return_ondemand) {	    
	    cmgr->returnActiveChannel(chan);
	    if (first_merge || v->first_merge) {
	      first_merge = 0; v->first_merge = 0;
	    }
	    chan = v->GetChan();
	  }
	  else {
	    cmgr->returnActiveChannel(v->GetChan());
	    if (first_merge || v->first_merge) {
	      first_merge = 0; v->first_merge = 0;
	    }
	  }

	  chan->setVS(this);
	  v->SetChan(NULL);
	  assert( ( (chan->getType() == OnDem) || (chan->getType() == Cont) ) );
  
	}
	// if both were in after_interaction the state is kept.
	if ((after_interaction && v->after_interaction)) after_interaction = 1;
	else after_interaction = 0;
      }
      else { // no stream in after_interaction

	assert(chan);

	// it is possible for the stream to have after_interaction == 0
	// and still has a "Contingency" type of channel. It happens
	// when 1 "after_interaction" stream merges with one 
	// "non-after_interaction" and return_ondemand == 1
	if ( ((chan->getType()==Cont)&&(!Config::return_ondemand)) ||
	     ((chan->getType() == OnDem)&&(Config::return_ondemand)) ) {

	  if (v->GetChan()) {
	    cmgr->returnActiveChannel(chan);
	    if (first_merge || v->first_merge) {
	      first_merge = 0; v->first_merge = 0;
	    }
	    chan = v->GetChan();
	  }
	}
	else {
	  
	  if (v->GetChan()) {
	    cmgr->returnActiveChannel(v->GetChan());
	    if (first_merge || v->first_merge) {
	      first_merge = 0; v->first_merge = 0;
	    }
	  }
	}
	
	assert(chan);
	chan->setVS(this);
	v->SetChan(NULL);
	assert( ( (chan->getType() == OnDem) || (chan->getType() == Cont) ) );
	
	if ( status == AD_INS ) { // 'this' is the stream that is in ADINS_STT
	  
	  // since 'this' virtstream should disappear - the convention is that
	  // the stream in ADINS_STT disappears in a join - the information on 
	  // the starting position position of the stream must be exchanged
	  MStream *tmp_mstr = v->mstr;
	  v->mstr = mstr;
	  mstr = tmp_mstr;
	  double tmpid = stream_id;
	  stream_id = v->stream_id;
	  v->stream_id = tmpid;
	  to_merge_with = v->to_merge_with;
	  
	  if (Config::join_info)
	    cout << "JOIN: " << k << " stream " << tmpid 
		 << " (this in AD_INS) invoked JOIN to stream "
		 << stream_id << endl;
	}
	else { // 'this' is the stream that is in PLAYING_STT

	  ad_to_go = v->ad_to_go;
	  time_to_go = v->time_to_go;
	  s_video_to_go = v->s_video_to_go;
	  video_to_go = v->video_to_go;
	  n_ad_time = v->n_ad_time;
	  continuous_video_time = v->continuous_video_time;

	}
      }
      v->mstr->end_position = position;
      v->mstr->end_time = k;
      mstrlist->Catenate(v->mstrlist);
      mstrlist->Append(v->mstr,0);
      
      // First add the number of streams v has in its "batch", then
      // add v, for if there was no aggregation, v would have remained
      // in the system
      n_batched += v->n_batched;
      n_batched +=1;
      
      if (Config::join_info)
	cout << "JOIN: " << k << " Streams " << stream_id << " (this) and " 
	     << v->StreamId() << " merged at position " << position << "\n";
    }
  }
  else {

    assert(chan);

    // if chan is of the type that should be returned...
    if ( ((chan->getType()==Cont)&&(!Config::return_ondemand)) ||
	 ((chan->getType() == OnDem)&&(Config::return_ondemand)) ) {

      // ... return it only if v has a valid channel ...
      if (v->GetChan()) {
	cmgr->returnActiveChannel(chan);
	if (first_merge || v->first_merge) {
	  first_merge = 0; v->first_merge = 0;
	}
	chan = v->GetChan();
      }
    }
    else { // else return the channel that v is holding

      if (v->GetChan()) {
	cmgr->returnActiveChannel(v->GetChan());
	if (first_merge || v->first_merge) {
	  first_merge = 0; v->first_merge = 0;
	}
      }
    }
    
    assert(chan);
    chan->setVS(this);
    v->SetChan(NULL);

    assert( ( (chan->getType() == OnDem) || (chan->getType() == Cont) ) );

    if (Config::join_info)
      cout << "JOIN: " << k << " Streams " << stream_id << " (this) with "
	   << NumUsers() << " and " 
	   << v->StreamId() << " with " << v->NumUsers() 
	   << " merged at position " << position << "\n";

  }
  
  Subptr *u = (Subptr *)v->users.Gethead();
  while(u)
    {
      if (u->sptr->GetUID() == Config::user_id) {
	cout << "User " << u->sptr->GetUID() << " transferring to stream "
	     << stream_id << endl;
      }
      u->sptr->SetMystream(this);
      u = (Subptr *)u->Getnext();
    }
  users.Catenate(&(v->users));

  delete v;
}

// reset statistics for continuous video time over the threshold
void Virtualstream::CheckVideoTime(double threshold)
{
  if (continuous_video_time >= threshold) {
    ad_to_go = Config::max_ad_time;
    time_to_go = Config::long_term_ad_time;
    n_ad_time = (int) Config::ratio_tprime_amax;
    video_to_go = -1.0;
    s_video_to_go = -1.0;
    continuous_video_time = -1.0;
  }
}

// advance the stream according to its state. Any interaction resets the 
// stream statistics and puts the stream in after_interaction state
void Virtualstream::Advancestream(double timenow, double adv)
{
  double prevad, prevtim, prevv, prevsv;
  int prevn;
  Status prevs;
  prevad = ad_to_go;
  prevtim = time_to_go;
  prevv = video_to_go;
  prevsv = s_video_to_go;
  prevn = n_ad_time;
  prevs = status;
  
  // must have a valid channel to advance unless it's bogus stream
  if (!bogus) assert(chan);

  switch(status) {
  case PLAYING: 
    if ( (config.data.AGG_POLICY == ADINSPOL) ) {
      // check for short term video time
      if (s_video_to_go >= 0.0) {
	s_video_to_go -= adv;
	// n_ad_time > 0 means that the short term constraint can
	// still be obeyed by adding ads. If it is zero, then no more
	// ads can be given until a new TPRIME window starts
	if ((n_ad_time > 0) && (s_video_to_go <= 0.0)) {
	  
	  // if ad_to_go > 0, then is should be fulfilling the
	  // requirement (time_to_go - ad_to_go) == m*AMAX, m integer
	  // By arbitrarily assigning ad_to_go to AMAX we might
	  // lose the synchronism, in which case it is better to avoid 
	  // such cases.
	  
	  double tmpdiffta = time_to_go - AMAX;
	  if ( (tmpdiffta >= 0) && (modulus(tmpdiffta,AMAX)==0)) {
	    ad_to_go = Config::max_ad_time;
	  }
	}
      }
      if (video_to_go >= 0.0) {
	video_to_go -= adv;
	if (video_to_go <= 0.0) {
	  // A new TPRIME window should start
	  ad_to_go = Config::max_ad_time;
	  time_to_go = Config::long_term_ad_time;
	  n_ad_time = (int) Config::ratio_tprime_amax;
	}
	else {
	  
	  // if continuous_video_time < 0, it means the stream
	  // doesn't need to keep track of which state it is in
	  // merely assume it has been playing video since -Infinite
	  // continuous_video is set to zero by an AD_INS state,
	  // after which it should be set to  -1 by CheckVideoTime
	  // after the stream has been in playing stt for a long stretch
	  // of time
	  
	  if ((continuous_video_time >= 0.0)) {
	    if (video_to_go < LAST_VIDEO_TIME) {
	      ad_to_go = time_to_go = 0.0;
	      n_ad_time = 0;
	    }
	    else {
	      double somevideo = video_to_go - LAST_VIDEO_TIME;
	      double restv = modulus(somevideo,VMIN);
	      int nrestv = (int)ceil(somevideo/VMIN);
	      
	      if (restv > 0) {
		assert(nrestv < Config::ratio_tprime_amax);
		ad_to_go = 0;
		time_to_go = nrestv*AMAX;
		n_ad_time = nrestv;
		if (s_video_to_go < 0) {
		  s_video_to_go = restv;
		}
	      }
	    }
	  }
	}
      }
      mstr->amount_in_play += adv;
      continuous_video_time += adv;
      CheckVideoTime(LAST_VIDEO_TIME);
      
      position +=adv;
      if(position >= mv->Duration())  {
	mstr->end_time = k;
	mstr->end_position = position;
	status = ATEOMV;
	newEvent = 1;
      }
    }
    else {
      position += adv;
      if (position >= mv->Duration()) {
	status = ATEOMV;
	newEvent = 1;
      }
    }
    break;
  case FASTPLAY: 
    position += config.data.FAST_PLAY_RATE * adv;
    if(position >= mv->Duration()) {
      status = ATEOMV;
      newEvent = 1;
    }
    break;

  case AD_INS:
    if ( (ad_to_go == Config::max_ad_time) && 
	 (time_to_go == Config::long_term_ad_time) ) {
      s_video_to_go = Config::min_video_time;
      video_to_go = (Config::long_term_video);
    }
    
    if ( (ad_to_go < adv) || (time_to_go < adv)) {
      cout << "\nERROR: stream " << stream_id 
	   << " violating QoS constraints\nad_to_go = " << ad_to_go
	   << " time_to_go = " << time_to_go << "\nadv = " << adv <<endl;
      Dumpdata();
      cout << "Debuggin stream commands: " << endl;
      global_ad_ins.debug_stream_id(StreamId());
      cout << "Debugging stream merging with stream " << StreamId() << endl;
      global_ad_ins.debug_merge_with(StreamId());
      assert(0);
    }
    
    ad_to_go -= adv;
    time_to_go -= adv;
    
    if (ad_to_go <= 0.0) {
      n_ad_time -= 1;
      
      if (n_ad_time > 0) {
	// still in one of the initial steps of a TPRIME window
	s_video_to_go = Config::min_video_time;
      }
      else {
	// in the end of TPRIME window. Trying to give more video time
	s_video_to_go = Config::long_term_video -
	  ((Config::ratio_tprime_amax-1)*VMIN);
      }
    }
    mstr->amount_in_ad += adv;
    continuous_video_time = 0.0;
    if (time_to_go < ad_to_go) {
      cout << "ERROR: stream " << StreamId() << " with time_to_go < ad_to_go\n";
      cout << "status = " << (int)status; print_Stts(status);
      cout << "ad_to_go = " << ad_to_go << "\ntime_to_go = "
	   << time_to_go << "\nvideo_to_go = " << video_to_go 
	   << "\ns_video_to_go = " << s_video_to_go << "\nn_ad_time = "
	   << n_ad_time << "\nmstr->amount_in_ad = " << mstr->amount_in_ad 
	   << endl;
      Virtualstream *dbvs = ((VSList *)mylist)->GetVS(\
						      global_ad_ins.debug_merge_with(StreamId()));
      if (dbvs)
	dbvs->Dumpdata();
      else 
	cout << "ERROR: not found stream to merge with " << StreamId() << endl;
      assert((time_to_go>=ad_to_go));
    }      
    break;

  case PAUSED: 
    if(until_time  <= timenow) {
      status = PLAYING;
      newEvent = 1;
      if ( (config.data.AGG_POLICY == ADINSPOL) ) {
	CheckVideoTime(-DBL_MAX);
	after_interaction = 1;
      }
    }
    break;
  case FWDING: 
    position += (config.data.FF_RATE*adv);
    if(position > mv->Duration()) position = mv->Duration();
    if(until_time  <= timenow) {
      status = PLAYING;
      newEvent = 1;
      if ( (config.data.AGG_POLICY == ADINSPOL) ) {
	CheckVideoTime(-DBL_MAX);
	after_interaction = 1;
      }
    }
    break;
  case RWDING: 
    position -= config.data.RW_RATE*adv;
    if(position < 0.) position = 0.;
    if(until_time  <= timenow) {
      status = PLAYING;
      newEvent = 1;
      if ( (config.data.AGG_POLICY == ADINSPOL) ) {
	CheckVideoTime(-DBL_MAX);
	after_interaction = 1;
      }
    }
    break;
  case ATEOMV:
    break;
  case INBATCH: 
    if(timenow >= until_time) status = RTOPLAY;
    break;
  case ADVACN: 
    break;
  default:
    break;
  }

  if (time_to_go < ad_to_go) {
    cout << "ERROR: stream " << StreamId() << " with time_to_go < ad_to_go\n";
    cout << "status = " << (int)status; print_Stts(status);
    cout << "ad_to_go = " << ad_to_go << "\ntime_to_go = "
	 << time_to_go << "\nvideo_to_go = " << video_to_go 
	 << "\ns_video_to_go = " << s_video_to_go << "\nn_ad_time = "
	 << n_ad_time << "\nmstr->amount_in_ad = " << mstr->amount_in_ad 
	 << endl;
    global_ad_ins.debug_stream_id(StreamId());
    assert((time_to_go>=ad_to_go));
  }      
  
  if (Config::plot_time_pos) {
    // using AMIN to diminish number of outputs
    //    if ( modulus(k,AMIN) == 0) 
    cout << " " << mv->MovieId() << "," 
	 << stream_id << "," 
	 << position << ","
	 << NumUsers() << " ";
  }
}

int Virtualstream::Matchstream(Virtualstream *v)
{
  if ((mv == v->mv) && (fabs(GetPos() - v->GetPos()) <= EPSILON)
      && (Getstatus() < PAUSED) && (v->Getstatus() < PAUSED))
    {
      int i = ((v->to_merge_with == stream_id) || (to_merge_with == v->StreamId()));

      if ( ((bogus) && !(i)) || ((v->bogus) && !(i))) {
	return FAILURE;
      }
      return SUCCESS;
    }
  else
    {
      return FAILURE;
    }
}


int VSList::Present(double str_id)
{
  Virtualstream *cursor = (Virtualstream *)head;
  while(cursor) 
    {
      if (cursor->stream_id==str_id)
	return (SUCCESS);
      cursor = (Virtualstream *)cursor->next;
    }
  
  return (FAILURE);
}

Virtualstream *VSList::GetVS(double str_id)
{
  Virtualstream *cursor = (Virtualstream *)head;
  while(cursor) 
    {
      if (cursor->stream_id==str_id)
	return (cursor);
      cursor = (Virtualstream *)cursor->next;
    }
  
  return (NULL);
}

Virtualstream *VSList::Findsubptrvs(Subscriber *u)
{
  Virtualstream *cursor = (Virtualstream *)head;
  while (cursor) {

    if ( (cursor->Findsubptr(u)) != NULL )
      return (cursor);
    cursor = (Virtualstream *)cursor->next;
  }
  return NULL;
}

Subptr *Virtualstream::Findsubptr(Subscriber *u)
{
  Subptr *tmp = (Subptr *) users.Gethead();
  while(tmp)
    {
      if(tmp->sptr == u) return(tmp);
      tmp = (Subptr *) tmp->Getnext();
    }
  return(NULL);
}

void Virtualstream::Dumpdata(void)
{
  cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;
  cout << "STREAM_ID = " << stream_id << "\t";
  cout << "Position = " << position << "\t";
  cout << "Until = " << until_time << "\n";
  cout << "Status = " << (int)status << " ";
  print_Stts(status);
  cout << " bogus = " << ((int)bogus) << " been_interaction = "
       << ((int)been_interaction) << " after_interaction = " 
       << ((int)after_interaction) << " to_merge_with " 
       << to_merge_with  
       << "\na:" << ad_to_go << " t:" << time_to_go 
       << " v:" << video_to_go << " sv:"<<s_video_to_go
       << " n:" << n_ad_time << endl;

  mv->Dumpdata();
  users.Dumpdata(); 	
}



Cost Virtualstream::GetCostBandwidth()
{
  Cost cost;
  double endp = position, endt = k;

  if (status == ATEOMV) {
    endp = mstr->end_position;
    endt = mstr->end_time;
  }

  MStream *cursor = (MStream *)mstrlist->Gethead();
  while (cursor) {
    cost.original += (endp - cursor->starting_position);
    cost.used += (cursor->end_time - cursor->starting_time);
    cost.amount_in_ad += (cursor->amount_in_ad);
    cost.amount_in_play += (cursor->amount_in_play);
    cursor = cursor->getNext();
  }
  
  cost.original += (endp- mstr->starting_position);
  cost.used += (endt - mstr->starting_time);
  cost.amount_in_ad += (mstr->amount_in_ad);
  cost.amount_in_play += (mstr->amount_in_play);
  return (cost) ;
}

void VSList::ResetAdTime()
{
  Virtualstream *cursor = (Virtualstream *)head;

  while (cursor) {
    cursor->Setstatus(PLAYING);
    cursor = (Virtualstream *)cursor->Getnext();
  }
}

void VSList::CheckReset()
{  
  Virtualstream *cursor=(Virtualstream *)head;

  while (cursor) {

    if ( cursor->ad_to_go < AMAX) {
      cout << "ResetAdTime: stream " << cursor->StreamId() 
	   << " ad_to_go not reset: " << cursor->ad_to_go << endl;
    }
    if ( cursor->time_to_go < TPRIME) {
      cout << "ResetAdTime: stream " << cursor->StreamId() 
	   << " time_to_go not reset: " << cursor->time_to_go << endl;
    }
    if ( cursor->s_video_to_go > 0.0){
      cout << "ResetAdTime: stream " << cursor->StreamId() 
	   << " s_video_to_go not reset: " << cursor->s_video_to_go << endl;
    }
    if (cursor->video_to_go > 0.0) {
      cout << "ResetAdTime: stream " << cursor->StreamId() 
	   << " video_to_go not reset: " << cursor->video_to_go << endl;
    }
    if (cursor->n_ad_time < (int)Config::ratio_tprime_amax) {
      cout << "ResetAdTime: stream " << cursor->StreamId() 
	   << " n_ad_time not reset: " << cursor->n_ad_time << endl;
    }
    cursor = (Virtualstream *)cursor->Getnext();
  }
  cout << "Done VSList->CheckReset\n";
}

int VSList::GetNBatched()
{
  return 0;
}

int VSList::GetNumberVS()
{
  Virtualstream *cursor = (Virtualstream *)head;
  int nvs = 0;

  while (cursor) {
    if (cursor->bogus == 0) 
      nvs += 1;
    cursor = (Virtualstream *)cursor->Getnext();
  }
  return nvs;
}

void VSList::UsersPerStream()
{
  Virtualstream *cursor = (Virtualstream *)head;
  
  if (Config::show_users_per_stream) {

    while (cursor) {
      if(cursor->bogus == 0) {
	cout << cursor->StreamId() << " - " << cursor->NumUsers() 
	     << " - " << cursor->GetPos();
	if (Config::debug_predict == 1) {
	  print_Stts(cursor->Getstatus());
	}
	cout <<  " : ";
      }
      cursor = (Virtualstream *)cursor->Getnext();
    }
  }
}

void VSList::GetCostBandwidth()
{
  cout
    << "\nCOST: VSL:\tamount_in_ad = " << (listcost.amount_in_ad)
    << "\nCOST: VSL:\tamount_in_play = " << (listcost.amount_in_play)
    << "\nCOST: VSL:\tad/play = " 
    << (listcost.amount_in_ad/listcost.amount_in_play)
    << "\nCOST: VSL:\tplay/ad = " 
    << (listcost.amount_in_play/listcost.amount_in_ad)
    << "\nCOST: VSL:\tad+play = "
    << (listcost.amount_in_play + listcost.amount_in_ad)
    << endl;

}


void VSList::Advanceall(double curtim)
{

  if (Config::plot_time_pos) {
    cout << "position at "<< curtim << " - ";
  }

  Virtualstream *cursor = (Virtualstream *)head;
  while(cursor) 
    {
      cursor->Advancestream(curtim, config.data.SIMGRAIN);
      cursor = (Virtualstream *)cursor->next;
    }

  if (Config::plot_time_pos) {
    cout << endl;
  }

}


Virtualstream *VSList::GetMatch(Virtualstream *v)
{
  Virtualstream *cursor = (Virtualstream *)head;
  while(cursor)
    {
      if ((cursor->Matchstream(v)==SUCCESS) && (v!=cursor))
	return (cursor);
      cursor = (Virtualstream *)cursor->next;
    }
  return(NULL);
}


void VSList::Compact(ChannelManager *cmgr)
{
  Virtualstream *tmp = NULL, *tmp2 = NULL, *tmp3 = NULL;
  
  if (num_nodes < 1) return;
  
  tmp = (Virtualstream *)head;
  tmp3 = (Virtualstream *)tmp->next;
  while (tmp)
    {
      if ((tmp2 = GetMatch(tmp)) != NULL)
	{
	  tmp3 = (Virtualstream *)tmp->next;
	  Detach(tmp);
	  tmp2->Joinstream(tmp, this, cmgr);
	  tmp = tmp3;
	}
      else tmp = (Virtualstream *)tmp->next;    
    }
}


// If stream is ready to play, launch the batch
void VSList::Launchbatch(VSList *b, double curtim)
{
  Virtualstream *cursor = (Virtualstream *)b->head;
  while(cursor) 
    {
      Virtualstream *tmp = cursor;
      cursor = (Virtualstream *)cursor->next;
      if(tmp->status==RTOPLAY) 
	{
	  tmp->status = PLAYING;
	  tmp->until_time = curtim + tmp->mv->Duration();
	  newEvent = 1;

	  Append(b->Detach(tmp), tail);
	  tmp->mstr->starting_time = curtim;
	}
    }
}

// If stream has exited, remove users from active subpool,
// return subscribers to idle pool and delete stream
void VSList::Handleexits(Subscriberlist *active, Subscriberlist *idle, ChannelManager *cmgr)
{
  Cost tmp_cost;
  int tmpNumUsers = 0;
  Virtualstream *cursor = (Virtualstream *)head;
  while(cursor) 
    {
      Virtualstream *tmp = cursor;
      cursor = (Virtualstream *)cursor->next;

      if(tmp->status==ATEOMV) 
	{

	  if (tmp->bogus == 0) {
	    tmp_cost = tmp->GetCostBandwidth();
	    listcost.original += tmp_cost.original;
	    listcost.used += tmp_cost.used;
	    listcost.amount_in_ad += tmp_cost.amount_in_ad;
	    listcost.amount_in_play += tmp_cost.amount_in_play;
	    tmpNumUsers = tmp->NumUsers();
	    Subptr *subcur = (Subptr *)(tmp->users.Gethead());
	    while(subcur)
	      {
		Subptr *subtmp = subcur;
		subcur = (Subptr *)subcur->Getnext();
		subtmp->sptr->SetRequest(PLAY);
		subtmp->sptr->SetDuration(DBL_MAX);
		idle->Prepend(active->Detach(subtmp->sptr), NULL);	
		tmp->users.Remove(subtmp);
	      }
	    assert(tmp->GetChan());
	    cmgr->returnActiveChannel(tmp->GetChan());
	    tmp->SetChan(NULL);
	
	  }
	  assert( (tmp->GetChan() == NULL) );

	  tmp->mstrlist->DeleteAllNodes();
	  Remove(tmp);
	}
    }
}

void VSList::RemoveBogus()
{
  Virtualstream *cursor = (Virtualstream *)head;
 
  while(cursor) {
    Virtualstream *tmp = cursor;
    cursor = (Virtualstream *)cursor->next;
    
    if(tmp->bogus==1) {
      Remove(tmp);
    }
  }
}


void VSList::ChangeStatus(double stream_id, Status status)
{
  Virtualstream *t;

  for (t = (Virtualstream *) head; t; t = (Virtualstream *) t->next)
    {
      if (t->StreamId() == stream_id)
	{
	  t->Setstatus(status);
	  return;
	}
    }
}


Status VSList::GetStat(double stream_id)
{
  Virtualstream *t;

  for (t = (Virtualstream *) head; t; t = (Virtualstream *) t->next)
    {
      if (t->StreamId() == stream_id)
	return t->Getstatus();
    }
  return UNKNOWN;
}

double VSList::GetPosition(double stream_id)
{
  Virtualstream *t;

  for (t = (Virtualstream *) head; t; t = (Virtualstream *) t->next)
    {
      if (t->StreamId() == stream_id)
	return t->GetPos();
    }
  return 0;  
}

void VSList::ChangeInList(STREAM *s, int index, NStatus newstat)
{ 
  double str_id = s[index].stream_id;
  Status stat = GetStat(str_id);
  switch (newstat) {
  case NORMAL:
    s[index].status = NORMAL;
    if ((Present(str_id)==SUCCESS) && (stat<PAUSED))
      ChangeStatus(str_id, PLAYING);
    break;
  case ACCEL:
    s[index].status = ACCEL;
    if ((Present(str_id)==SUCCESS) && (stat<PAUSED))
      ChangeStatus(str_id, FASTPLAY);
    break;

  case NSADINS:
    s[index].status = NSADINS;
    if ((Present(str_id) == SUCCESS) && (stat<PAUSED))
      ChangeStatus(str_id, AD_INS);
    break;

  default: break;
  }
}


// Executes the GREEDY merging policy within a given WINDOW
void VSList::GreedyMerge(STREAM *streams, int nstreams, double *window)
{
  int i;

  if (streams[0].status == ACCEL)
    ChangeInList(streams, 0, NORMAL);

  for (i=0; i<nstreams-1; i++) 
    {
      if (!CARRY_SAME_PROGRAM(streams[i].movie_id, streams[i+1].movie_id))
	{
	  if (streams[i+1].status == ACCEL) // If the leading stream is accel. then slow it down
	    ChangeInList(streams, i+1, NORMAL);
	  continue;
	}

      if (fabs(streams[i].position - streams[i+1].position) <= EPSILON)
	{
	  ChangeInList(streams, i, NORMAL);
	  ChangeInList(streams, i+1, NORMAL);
	  continue;
	}
    
      if (streams[i].status == ACCEL)
	ChangeInList(streams, i+1, NORMAL);
      else {
	if (Kfactor*(streams[i].position-streams[i+1].position) > (*window)) 
	  ChangeInList(streams, i+1, NORMAL);
	else
	  ChangeInList(streams, i+1, ACCEL);
      }
    }
  (*window) -= 0.0333; // Decrement window by 33.3 ms (1 frame equiv)

  if (*window <= EPSILON)
    {
      for (i=0; i<nstreams; i++)
	ChangeInList(streams, i, NORMAL);
    }

}




// Run DSAP and celebrate ....... Yahoooo !!!

// if recomp = 1 then the clustering engine is called and all the streams
// have their "directions" set. If recomp = 0, then all the streams
// states (FASTPLAY, PLAY, AD_INS, etc) are updated according to the
// simulation time

void VSList::Aggregate(int policy, char recomp, double *window, double curtim)
{
  int i;
  Policy p;
  char greedy = 0;
  Virtualstream *t;

  switch(policy) 
    {  
    case 0: p = EMCLRSMA; break;
    case 1: p = EMCLAFL; break;
    case 2: p = GREEDY; greedy = 1; break;
    case 3: p = EMCLRSMAHEUR; break;
    case 4: p = RSMA; break;
    case 5: p = RSMAHEUR; break;
    case 6: p = ADINSPOL; break;
    default : break;
    }

  // If recompute is set, create array of STREAMS from VSList
  // and then invoke cluster computation

  if(recomp || greedy)
    {
      numstrms = 0;
      if(strms) {free(strms); strms = NULL;}

      strms = (STREAM *) malloc(sizeof(STREAM) * num_nodes); 
      
      // Build new array of streams which are PLAYING/FASTPLAY
      for(t=(Virtualstream *)head; t; t=(Virtualstream *)t->next)
	{
	  if ((int)t->newArrival) {
	    Channel *tmpChan = t->GetChan();
	    assert(tmpChan);
	    assert( ( (tmpChan->assignedAt > 0.0) && (tmpChan->assignedAt <= k) ));
	    totalTTR += (k - tmpChan->assignedAt);
	    nStreamsTTR += 1.0;
	    t->newArrival = 0;
	  }

	  // obtain the recomputation time and reset first_merge for some streams
	  t->last_recomp = k;

	  if(t->status == PLAYING)
	    {
	      t->first_merge = 1;
	
	      t->after_interaction = 0;
	      t->been_interaction = 0;
	      t->to_merge_with = -1.0;
	      strms[numstrms].movie_id = t->mv->MovieId();
	      strms[numstrms].stream_id = t->stream_id;
	      strms[numstrms].position = t->position;
	      strms[numstrms].status = NORMAL;
	
	      strms[numstrms].movie_length = t->mv->Duration();
	      strms[numstrms].ad_to_go = t->ad_to_go;
	      strms[numstrms].time_to_go = t->time_to_go;
	      strms[numstrms].video_to_go = t->video_to_go;
	
	      numstrms++;
	      continue;
	    }

	  if(t->status == FASTPLAY)
	    {
	      strms[numstrms].movie_id = t->mv->MovieId();
	      strms[numstrms].stream_id = t->stream_id;
	      strms[numstrms].position = t->position;
	      strms[numstrms].status = ACCEL;
	      strms[numstrms].movie_length = t->mv->Duration();
	      strms[numstrms].ad_to_go = t->ad_to_go;
	      strms[numstrms].time_to_go = t->time_to_go;
	      strms[numstrms].video_to_go = t->video_to_go;
	
	      numstrms++;
	      continue;
	    }
      
	  if (t->status == AD_INS)
	    {
	      t->first_merge = 1;
	  
	      t->after_interaction = 0;
	      t->been_interaction = 0;
	      t->to_merge_with = -1.0;
	      strms[numstrms].movie_id = t->mv->MovieId();
	      strms[numstrms].stream_id = t->stream_id;
	      strms[numstrms].position = t->position;
	      strms[numstrms].status = NSADINS;
	      strms[numstrms].movie_length = t->mv->Duration();
	      strms[numstrms].ad_to_go = t->ad_to_go;
	      strms[numstrms].time_to_go = t->time_to_go;
	      strms[numstrms].video_to_go = t->video_to_go;
	      numstrms++;
	      continue;
	    }
	}
    
      // Now we have the streams in the standard structure format
      // Sort streams by movie id and decreasing position
      if (greedy)
	{
	  qsort(strms, numstrms, sizeof(STREAM), compar);
	  GreedyMerge(strms, numstrms, window);
	  return;
	}
      // Pass to invoke_policy for all policies 
      // other than GREEDY
    
      if (numstrms)
	{
	  /* Call the Clustering Algorithms on the Stream Array */
	  invoke_policy(p, strms, numstrms, *window, curtim);
	  if (Config::plot_time_pos)
	    cout << "Recomputation invoked at time " << curtim << endl;
	
	  // For each element in VSList, get_status and set_status 
	  for(i=0; i<numstrms; i++)
	    {
	      double str_id = strms[i].stream_id;
	      double pos = strms[i].position;
	      Virtualstream *tvs = GetVS(str_id);
	    
	      if ( tvs ) {
		if ( (config.data.AGG_POLICY == ADINSPOL) ) {
		  tvs->to_merge_with = global_ad_ins.to_merge_with(str_id);
		  if (tvs->to_merge_with == STREAM_NOT_FOUND) {
		    cout << "ERROR: stream " << str_id << " not found in clustering engine\n";
		    assert(0);
		  }
		}
	      
		switch(get_stream_status(str_id,curtim,pos,p)) {
		case NORMAL:
		  strms[i].status = NORMAL;
		  tvs->Setstatus(PLAYING);
		  break;
		case ACCEL:
		  strms[i].status = ACCEL;
		  ChangeStatus(str_id, FASTPLAY);
		  break;
		case NSADINS:
		  strms[i].status = NSADINS;
		  tvs->Setstatus(AD_INS);
		  break;
		default: break;
		}
	      }
	      else {
		cerr << "ERROR: stream " << str_id 
		     << " not found in VSList right after recomputation \n";
		cout << "ERROR: stream " << str_id 
		     << " not found in VSList right after recomputation \n";
	      }
	    }
	}
    }
  else 
    {
      for(i=0; i<numstrms; i++)
	{
	  double str_id = strms[i].stream_id;
	  double pos = GetPosition(str_id);
	  Status stat = GetStat(str_id);
	  
	  if ((Present(str_id)==SUCCESS) && (stat<PAUSED)) 
	    {
	      
	      Virtualstream *testvs = GetVS(str_id);
	      if ( testvs->been_interaction == 0) {
		switch(get_stream_status(str_id, curtim, pos, p))
		  {
		  case NORMAL: 
		    strms[i].status = NORMAL; 
		    ChangeStatus(str_id, PLAYING);
		    break;
		  case ACCEL: 
		    strms[i].status = ACCEL; 
		    ChangeStatus(str_id, FASTPLAY);	    
		    break;
		  case NSADINS:
		    strms[i].status = NSADINS;
		    ChangeStatus(str_id, AD_INS);
		    break;
		  default: break;
		  }
	      }
	      else {
		assert( (config.data.AGG_POLICY == ADINSPOL) );
	      }
	    }
	}
    }
}

// End of File


